from .VObject import VObject
from .VSignal import VSignal

class VCoreApplication(VObject):
    """
    Core application class. Only one instance is allowed to exist.
    """
    vApp = None

    def __init__(self, argv):
        super().__init__()
        self._timers = []

        if VCoreApplication.vApp is not None:
            raise Exception("Only one application is allowed")

        VCoreApplication.vApp = self
        self.aboutToQuit = VSignal(self)

    def addTimer(self, timer):
        """
        Add a timer to the application.
        This routine should not be called manually.
        """
        self._timers.append(timer)

    def exit(self, retcode=0):
        """
        Exits the application.
        """
        VCoreApplication.vApp = None

    def sendEvent(self, receiver, event):
        """
        Directly send an event to a receiver.
        """
        receiver.event(event)

    def applicationName(self):
        raise NotImplementedError()

    def applicationVersion(self):
        raise NotImplementedError()

    def instance(self):
        raise NotImplementedError()

    def exec_(self):
        raise NotImplementedError()

    def processEvents(self, flags):
        raise NotImplementedError()

    def postEvent(self, receiver, event):
        raise NotImplementedError()

    def sendPostedEvents(self, receiver, event_type):
        raise NotImplementedError()

    def removePostedEvents(self, receiver, event_type):
        raise NotImplementedError()

    def hasPendingEvents(self):
        raise NotImplementedError()

    def notify(self, receiver, event):
        raise NotImplementedError()

    def applicationDirPath(self):
        raise NotImplementedError()

    def applicationFilePath(self):
        raise NotImplementedError()

    def startingUp(self):
        raise NotImplementedError()

    def closingDown(self):
        raise NotImplementedError()

    def setEventFilter(self, event_filter):
        raise NotImplementedError()






from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class InsertLineAfterCommand(BufferCommand):
    """
    Command to perform insertion of a line after the cursor current line
    Moves the cursor to the first non-blank character of the newly added line
    """
    def __init__(self, buffer, text):
        super().__init__(buffer)
        self._text = text

    def execute(self):
        document = self._buffer.document
        cursor = self._buffer.cursor

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()
        cursor.toPos(pos)

        document.insertLine(pos[0]+1, self._text)
        document.lineMetaInfo("Change").setData("added", pos[0]+1)
        cursor.toCharFirstNonBlankForLine(pos[0]+1)
        return CommandResult(True, None)

    def undo(self):
        self._buffer.document.deleteLine(self.savedCursorPos()[0]+1)
        self.restoreCursorPos()




from . import VColor
from ..consts import Index
import curses, _curses
import select
import sys
import os
import logging
import threading

class VException(Exception): pass

class VScreen(object):
    def __init__(self):
        # Timeout so that ncurses sends out the pure esc key instead of considering it
        # the start of a escape command. We need this to exit insert mode in vai, and
        # it makes sense overall
        os.environ["ESCDELAY"] = "25"
        try:
            self._curses_screen = curses.initscr()
        except:
            raise VException("Cannot initialize screen")
        curses.start_color()
        curses.use_default_colors()
        curses.noecho()
        curses.cbreak()
        curses.raw()
        self._curses_screen.keypad(1)
        self._curses_screen.nodelay(True)
        self._curses_screen.leaveok(True)
        self._curses_screen.notimeout(True)

        # ncurses is not thread safe, we need a lock to prevent writing collide with reading
        self._curses_lock = threading.Lock()

        # Resolves the rgb color to the screen color
        self._color_lookup_cache = {}

        # Resolves fg, bg native index to the associated color pair index
        self._attr_lookup_cache = {}

        # The first color pair is always defined with index 0 and contains the default fg and bg colors
        self._color_pairs = [ (-1, -1) ]

        self._cursor_pos = (0,0)
        self.logger = logging.getLogger(self.__class__.__name__)
        if hasattr(self, "debug"):
            self.logger.setLevel(self.debug)
        else:
            self.logger.setLevel(logging.CRITICAL+1)
        VGlobalScreenColor.init(self.numColors())

    def reset(self):
        self._curses_screen.keypad(0)
        curses.echo()
        curses.nocbreak()
        curses.endwin()

    def refresh(self):
        with self._curses_lock:
            self._curses_screen.noutrefresh()
            curses.setsyx(self._cursor_pos[Index.Y], self._cursor_pos[Index.X])
            curses.doupdate()

    def rect(self):
        return self.topLeft() + self.size()

    def size(self):
        with self._curses_lock:
            h, w = self._curses_screen.getmaxyx()
        return (w,h)

    def topLeft(self):
        return (0,0)

    def width(self):
        return self.size()[Index.SIZE_WIDTH]

    def height(self):
        return self.size()[Index.SIZE_HEIGHT]

    def getKeyCode(self):
        # Prevent to hold the GIL
        select.select([sys.stdin], [], [])

        with self._curses_lock:
            c = self._curses_screen.getch()
            # 27 = Escape character, but it's also used to create multikeys.
            # So we detect a pure escape by getting again. If it returns
            # -1, it means there's no additional keys and it's a pure escape
            if c == 27:
                next_c = self._curses_screen.getch()
                if next_c == -1:
                    pass
                # FIXME later: handle multikeys

        return c

    def write(self, pos, string, fg_color=None, bg_color=None):
        x,y = pos
        w,h = self.size()

        out_string = string

        if y < 0 or y >= h or x >= w:
            self.logger.error("Out of bound in VScreen.write: pos=%s size=%s len=%d '%s'" % (str(pos), str(self.size()), len(string), string))
            return

        out_string = out_string[:w-x]

        if x < 0:
            self.logger.error("Out of bound in VScreen.write: pos=%s size=%s len=%d '%s'" % (str(pos), str(self.size()), len(string), string))
            out_string = string[-x:]

        if len(out_string) == 0:
            return

        attr = self.getColorAttributeCode(fg_color, bg_color)
        if (x+len(out_string) > w):
            self.logger.error("Out of bound in VScreen.write: pos=%s size=%s len=%d '%s'" % (str(pos), str(self.size()), len(string), string))
            out_string = out_string[:w-x]

        if (x+len(out_string) == w):
            with self._curses_lock:
                # Old ncurses trick. We can't write the very last character, so
                # we add everything but the first, and then push everything forward
                self._curses_screen.addstr(y, x, out_string[1:], attr)
                self._curses_screen.insstr(y, x, out_string[0], attr)
                self._curses_screen.noutrefresh()
        else:
            with self._curses_lock:
                self._curses_screen.addstr(y, x, out_string, attr)
                self._curses_screen.noutrefresh()

    def setColors(self, pos, colors):
        """
        Sets the color attributes for a specific line, starting at pos and forward until the colors
        array runs out.

        """

        x,y = pos
        w,h = self.size()

        out_colors = colors

        if y < 0 or y >= h or x >= w:
            self.logger.error("Out of bound in VScreen.setColors: pos=%s size=%s len=%d" % (str(pos), str(self.size()), len(colors)))
            return

        out_colors = out_colors[:w-x]

        if x < 0:
            self.logger.error("Out of bound in VScreen.setColors: pos=%s size=%s len=%d" % (str(pos), str(self.size()), len(colors)))
            out_colors = colors[-x:]

        if len(out_colors) == 0:
            return

        if (x+len(out_colors) > w):
            self.logger.error("Out of bound in VScreen.setColors: pos=%s size=%s len=%d" % (str(pos), str(self.size()), len(colors)))
            out_colors = out_colors[:w-x]

        for num, col in enumerate(out_colors):
            if len(col) == 1:
                fg_color = col[0]
                bg_color = None
                fg_font = None
            elif len(col) == 2:
                fg_color = col[0]
                fg_font = None
                bg_color = col[1]
            else:
                fg_color = col[0]
                fg_font = None
                bg_color = col[2]

            attr = self.getColorAttributeCode(fg_color, bg_color)
            with self._curses_lock:
                self._curses_screen.chgat(y, x+num, 1, attr)

    def numColors(self):
        return curses.COLORS

    def getColorAttributeCode(self, fg=None, bg=None):
        """Given fg and bg colors, find and return the correct attribute code to apply with chgat"""
        fg_screen = None if fg is None else self._findClosestColor(fg)
        bg_screen = None if bg is None else self._findClosestColor(bg)

        if (fg_screen, bg_screen) in self._attr_lookup_cache:
            return self._attr_lookup_cache[(fg_screen, bg_screen)]

        fg_index = -1 if fg_screen is None else fg_screen.colorIdx()
        bg_index = -1 if bg_screen is None else bg_screen.colorIdx()

        attr = self._getPairAttrFromColors(fg_index, bg_index)

        if fg_screen and fg_screen.attr():
            attr |= fg_screen.attr()

        self._attr_lookup_cache[(fg_screen, bg_screen)] = attr
        return attr

    def _getPairAttrFromColors(self, fg_index, bg_index):
        t = (fg_index, bg_index)

        if t in self._color_pairs:
            pair_index = self._color_pairs.index( (fg_index, bg_index) )
        else:
            pair_index = len(self._color_pairs)
            with self._curses_lock:
                curses.init_pair(pair_index, fg_index, bg_index)
            self._color_pairs.append((fg_index, bg_index))

        with self._curses_lock:
            attr = curses.color_pair(pair_index)

        return attr
        # Init color pairs

    def setCursorPos(self, pos):
        if self.outOfBounds(pos):
            self.logger.error("out of bound in Screen.setCursorPos: %s" % str(pos))
            return

        self._cursor_pos = pos

    def cursorPos(self):
        return self._cursor_pos

    def _findClosestColor(self, color):
        screen_color = self._color_lookup_cache.get(color.rgb)
        if screen_color is not None:
            return screen_color

        closest = sorted([(VColor.VColor.distance(color, screen_color),
                           screen_color)
                           for index, screen_color in enumerate(VGlobalScreenColor.allColors())
                         ],
                         key=lambda x: x[0]
                         )[0]
        self._color_lookup_cache[color.rgb] = closest[1]
        return closest[1]

    def outOfBounds(self, pos):
        x, y = pos
        return (x >= self.width() or y >= self.height() or x < 0 or y < 0)

class VScreenColor(object):
    def __init__(self, color_idx, attr, equiv_rgb):
        self._color_idx = color_idx
        self._attr = attr
        self._equiv_rgb = equiv_rgb

    def attr(self):
        return self._attr

    def colorIdx(self):
        return self._color_idx

    def equivRgb(self):
        return self._equiv_rgb

    @property
    def r(self):
        return self._equiv_rgb[0]
    @property
    def g(self):
        return self._equiv_rgb[1]
    @property
    def b(self):
        return self._equiv_rgb[2]

class VGlobalScreenColor(object):
    @classmethod
    def init(cls, num_colors):
        if num_colors == 8 or num_colors == 256:
            cls.term_0        = VScreenColor(0  , None, (0x00,0x00,0x00))
            cls.term_1        = VScreenColor(1  , None, (0x80,0x00,0x00))
            cls.term_2        = VScreenColor(2  , None, (0x00,0x80,0x00))
            cls.term_3        = VScreenColor(3  , None, (0x80,0x80,0x00))
            cls.term_4        = VScreenColor(4  , None, (0x00,0x00,0x80))
            cls.term_5        = VScreenColor(5  , None, (0x80,0x00,0x80))
            cls.term_6        = VScreenColor(6  , None, (0x00,0x80,0x80))
            cls.term_6        = VScreenColor(7  , None, (0xc0,0xc0,0xc0))
            cls.term_8        = VScreenColor(8  , None, (0x80,0x80,0x80))

            cls.black        = VScreenColor(curses.COLOR_BLACK, None, (0,0,0))
            cls.darkred      = VScreenColor(curses.COLOR_RED, None, (170,0,0))
            cls.darkgreen    = VScreenColor(curses.COLOR_GREEN, None, (0, 170, 0))
            cls.brown        = VScreenColor(curses.COLOR_YELLOW, None, (170, 170, 0))
            cls.darkblue     = VScreenColor(curses.COLOR_BLUE, None, (0, 0, 170))
            cls.darkmagenta  = VScreenColor(curses.COLOR_MAGENTA, None, (170, 0, 170))
            cls.darkcyan     = VScreenColor(curses.COLOR_CYAN, None, (0, 170, 170))
            cls.lightgray    = VScreenColor(curses.COLOR_WHITE, None, (170, 170, 170))

            cls.darkgray     = VScreenColor(curses.COLOR_BLACK, curses.A_BOLD, (100, 100, 100))
            cls.lightred     = VScreenColor(curses.COLOR_RED, curses.A_BOLD, (255,0,0))
            cls.lightgreen   = VScreenColor(curses.COLOR_GREEN, curses.A_BOLD, (0, 255, 0))
            cls.yellow       = VScreenColor(curses.COLOR_YELLOW, curses.A_BOLD, (255, 255, 0))
            cls.lightblue    = VScreenColor(curses.COLOR_BLUE, curses.A_BOLD, (0, 0, 255))
            cls.lightmagenta = VScreenColor(curses.COLOR_MAGENTA, curses.A_BOLD, (255, 0, 255))
            cls.lightcyan    = VScreenColor(curses.COLOR_CYAN, curses.A_BOLD, (0, 255, 255))
            cls.white        = VScreenColor(curses.COLOR_WHITE, curses.A_BOLD, (255,255,255))

            cls.red          = cls.lightred
            cls.green        = cls.lightgreen
            cls.blue         = cls.lightblue
            cls.magenta      = cls.lightmagenta
            cls.cyan         = cls.lightcyan
            cls.gray         = cls.lightgray

        if num_colors == 256:
            cls.term_9        = VScreenColor(9  , None, (0xff,0x00,0x00))
            cls.term_10       = VScreenColor(10 , None, (0x00,0xff,0x00))
            cls.term_11       = VScreenColor(11 , None, (0xff,0xff,0x00))
            cls.term_12       = VScreenColor(12 , None, (0x00,0x00,0xff))
            cls.term_13       = VScreenColor(13 , None, (0xff,0x00,0xff))
            cls.term_14       = VScreenColor(14 , None, (0x00,0xff,0xff))
            cls.term_15       = VScreenColor(15 , None, (0xff,0xff,0xff))
            cls.term_16       = VScreenColor(16 , None, (0x00,0x00,0x00))
            cls.term_17       = VScreenColor(17 , None, (0x00,0x00,0x5f))
            cls.term_18       = VScreenColor(18 , None, (0x00,0x00,0x87))
            cls.term_19       = VScreenColor(19 , None, (0x00,0x00,0xaf))
            cls.term_20       = VScreenColor(20 , None, (0x00,0x00,0xd7))
            cls.term_21       = VScreenColor(21 , None, (0x00,0x00,0xff))
            cls.term_22       = VScreenColor(22 , None, (0x00,0x5f,0x00))
            cls.term_23       = VScreenColor(23 , None, (0x00,0x5f,0x5f))
            cls.term_24       = VScreenColor(24 , None, (0x00,0x5f,0x87))
            cls.term_25       = VScreenColor(25 , None, (0x00,0x5f,0xaf))
            cls.term_26       = VScreenColor(26 , None, (0x00,0x5f,0xd7))
            cls.term_27       = VScreenColor(27 , None, (0x00,0x5f,0xff))
            cls.term_28       = VScreenColor(28 , None, (0x00,0x87,0x00))
            cls.term_29       = VScreenColor(29 , None, (0x00,0x87,0x5f))
            cls.term_30       = VScreenColor(30 , None, (0x00,0x87,0x87))
            cls.term_31       = VScreenColor(31 , None, (0x00,0x87,0xaf))
            cls.term_32       = VScreenColor(32 , None, (0x00,0x87,0xd7))
            cls.term_33       = VScreenColor(33 , None, (0x00,0x87,0xff))
            cls.term_34       = VScreenColor(34 , None, (0x00,0xaf,0x00))
            cls.term_35       = VScreenColor(35 , None, (0x00,0xaf,0x5f))
            cls.term_36       = VScreenColor(36 , None, (0x00,0xaf,0x87))
            cls.term_37       = VScreenColor(37 , None, (0x00,0xaf,0xaf))
            cls.term_38       = VScreenColor(38 , None, (0x00,0xaf,0xd7))
            cls.term_39       = VScreenColor(39 , None, (0x00,0xaf,0xff))
            cls.term_40       = VScreenColor(40 , None, (0x00,0xd7,0x00))
            cls.term_41       = VScreenColor(41 , None, (0x00,0xd7,0x5f))
            cls.term_42       = VScreenColor(42 , None, (0x00,0xd7,0x87))
            cls.term_43       = VScreenColor(43 , None, (0x00,0xd7,0xaf))
            cls.term_44       = VScreenColor(44 , None, (0x00,0xd7,0xd7))
            cls.term_45       = VScreenColor(45 , None, (0x00,0xd7,0xff))
            cls.term_46       = VScreenColor(46 , None, (0x00,0xff,0x00))
            cls.term_47       = VScreenColor(47 , None, (0x00,0xff,0x5f))
            cls.term_48       = VScreenColor(48 , None, (0x00,0xff,0x87))
            cls.term_49       = VScreenColor(49 , None, (0x00,0xff,0xaf))
            cls.term_50       = VScreenColor(50 , None, (0x00,0xff,0xd7))
            cls.term_51       = VScreenColor(51 , None, (0x00,0xff,0xff))
            cls.term_52       = VScreenColor(52 , None, (0x5f,0x00,0x00))
            cls.term_53       = VScreenColor(53 , None, (0x5f,0x00,0x5f))
            cls.term_54       = VScreenColor(54 , None, (0x5f,0x00,0x87))
            cls.term_55       = VScreenColor(55 , None, (0x5f,0x00,0xaf))
            cls.term_56       = VScreenColor(56 , None, (0x5f,0x00,0xd7))
            cls.term_57       = VScreenColor(57 , None, (0x5f,0x00,0xff))
            cls.term_58       = VScreenColor(58 , None, (0x5f,0x5f,0x00))
            cls.term_59       = VScreenColor(59 , None, (0x5f,0x5f,0x5f))
            cls.term_60       = VScreenColor(60 , None, (0x5f,0x5f,0x87))
            cls.term_61       = VScreenColor(61 , None, (0x5f,0x5f,0xaf))
            cls.term_62       = VScreenColor(62 , None, (0x5f,0x5f,0xd7))
            cls.term_63       = VScreenColor(63 , None, (0x5f,0x5f,0xff))
            cls.term_64       = VScreenColor(64 , None, (0x5f,0x87,0x00))
            cls.term_65       = VScreenColor(65 , None, (0x5f,0x87,0x5f))
            cls.term_66       = VScreenColor(66 , None, (0x5f,0x87,0x87))
            cls.term_67       = VScreenColor(67 , None, (0x5f,0x87,0xaf))
            cls.term_68       = VScreenColor(68 , None, (0x5f,0x87,0xd7))
            cls.term_69       = VScreenColor(69 , None, (0x5f,0x87,0xff))
            cls.term_70       = VScreenColor(70 , None, (0x5f,0xaf,0x00))
            cls.term_71       = VScreenColor(71 , None, (0x5f,0xaf,0x5f))
            cls.term_72       = VScreenColor(72 , None, (0x5f,0xaf,0x87))
            cls.term_73       = VScreenColor(73 , None, (0x5f,0xaf,0xaf))
            cls.term_74       = VScreenColor(74 , None, (0x5f,0xaf,0xd7))
            cls.term_75       = VScreenColor(75 , None, (0x5f,0xaf,0xff))
            cls.term_76       = VScreenColor(76 , None, (0x5f,0xd7,0x00))
            cls.term_77       = VScreenColor(77 , None, (0x5f,0xd7,0x5f))
            cls.term_78       = VScreenColor(78 , None, (0x5f,0xd7,0x87))
            cls.term_79       = VScreenColor(79 , None, (0x5f,0xd7,0xaf))
            cls.term_80       = VScreenColor(80 , None, (0x5f,0xd7,0xd7))
            cls.term_81       = VScreenColor(81 , None, (0x5f,0xd7,0xff))
            cls.term_82       = VScreenColor(82 , None, (0x5f,0xff,0x00))
            cls.term_83       = VScreenColor(83 , None, (0x5f,0xff,0x5f))
            cls.term_84       = VScreenColor(84 , None, (0x5f,0xff,0x87))
            cls.term_85       = VScreenColor(85 , None, (0x5f,0xff,0xaf))
            cls.term_86       = VScreenColor(86 , None, (0x5f,0xff,0xd7))
            cls.term_87       = VScreenColor(87 , None, (0x5f,0xff,0xff))
            cls.term_88       = VScreenColor(88 , None, (0x87,0x00,0x00))
            cls.term_89       = VScreenColor(89 , None, (0x87,0x00,0x5f))
            cls.term_90       = VScreenColor(90 , None, (0x87,0x00,0x87))
            cls.term_91       = VScreenColor(91 , None, (0x87,0x00,0xaf))
            cls.term_92       = VScreenColor(92 , None, (0x87,0x00,0xd7))
            cls.term_93       = VScreenColor(93 , None, (0x87,0x00,0xff))
            cls.term_94       = VScreenColor(94 , None, (0x87,0x5f,0x00))
            cls.term_95       = VScreenColor(95 , None, (0x87,0x5f,0x5f))
            cls.term_96       = VScreenColor(96 , None, (0x87,0x5f,0x87))
            cls.term_97       = VScreenColor(97 , None, (0x87,0x5f,0xaf))
            cls.term_98       = VScreenColor(98 , None, (0x87,0x5f,0xd7))
            cls.term_99       = VScreenColor(99 , None, (0x87,0x5f,0xff))
            cls.term_100      = VScreenColor(100, None, (0x87,0x87,0x00))
            cls.term_101      = VScreenColor(101, None, (0x87,0x87,0x5f))
            cls.term_102      = VScreenColor(102, None, (0x87,0x87,0x87))
            cls.term_103      = VScreenColor(103, None, (0x87,0x87,0xaf))
            cls.term_104      = VScreenColor(104, None, (0x87,0x87,0xd7))
            cls.term_105      = VScreenColor(105, None, (0x87,0x87,0xff))
            cls.term_106      = VScreenColor(106, None, (0x87,0xaf,0x00))
            cls.term_107      = VScreenColor(107, None, (0x87,0xaf,0x5f))
            cls.term_108      = VScreenColor(108, None, (0x87,0xaf,0x87))
            cls.term_109      = VScreenColor(109, None, (0x87,0xaf,0xaf))
            cls.term_110      = VScreenColor(110, None, (0x87,0xaf,0xd7))
            cls.term_111      = VScreenColor(111, None, (0x87,0xaf,0xff))
            cls.term_112      = VScreenColor(112, None, (0x87,0xd7,0x00))
            cls.term_113      = VScreenColor(113, None, (0x87,0xd7,0x5f))
            cls.term_114      = VScreenColor(114, None, (0x87,0xd7,0x87))
            cls.term_115      = VScreenColor(115, None, (0x87,0xd7,0xaf))
            cls.term_116      = VScreenColor(116, None, (0x87,0xd7,0xd7))
            cls.term_117      = VScreenColor(117, None, (0x87,0xd7,0xff))
            cls.term_118      = VScreenColor(118, None, (0x87,0xff,0x00))
            cls.term_119      = VScreenColor(119, None, (0x87,0xff,0x5f))
            cls.term_120      = VScreenColor(120, None, (0x87,0xff,0x87))
            cls.term_121      = VScreenColor(121, None, (0x87,0xff,0xaf))
            cls.term_122      = VScreenColor(122, None, (0x87,0xff,0xd7))
            cls.term_123      = VScreenColor(123, None, (0x87,0xff,0xff))
            cls.term_124      = VScreenColor(124, None, (0xaf,0x00,0x00))
            cls.term_125      = VScreenColor(125, None, (0xaf,0x00,0x5f))
            cls.term_126      = VScreenColor(126, None, (0xaf,0x00,0x87))
            cls.term_127      = VScreenColor(127, None, (0xaf,0x00,0xaf))
            cls.term_128      = VScreenColor(128, None, (0xaf,0x00,0xd7))
            cls.term_129      = VScreenColor(129, None, (0xaf,0x00,0xff))
            cls.term_130      = VScreenColor(130, None, (0xaf,0x5f,0x00))
            cls.term_131      = VScreenColor(131, None, (0xaf,0x5f,0x5f))
            cls.term_132      = VScreenColor(132, None, (0xaf,0x5f,0x87))
            cls.term_133      = VScreenColor(133, None, (0xaf,0x5f,0xaf))
            cls.term_134      = VScreenColor(134, None, (0xaf,0x5f,0xd7))
            cls.term_135      = VScreenColor(135, None, (0xaf,0x5f,0xff))
            cls.term_136      = VScreenColor(136, None, (0xaf,0x87,0x00))
            cls.term_137      = VScreenColor(137, None, (0xaf,0x87,0x5f))
            cls.term_138      = VScreenColor(138, None, (0xaf,0x87,0x87))
            cls.term_139      = VScreenColor(139, None, (0xaf,0x87,0xaf))
            cls.term_140      = VScreenColor(140, None, (0xaf,0x87,0xd7))
            cls.term_141      = VScreenColor(141, None, (0xaf,0x87,0xff))
            cls.term_142      = VScreenColor(142, None, (0xaf,0xaf,0x00))
            cls.term_143      = VScreenColor(143, None, (0xaf,0xaf,0x5f))
            cls.term_144      = VScreenColor(144, None, (0xaf,0xaf,0x87))
            cls.term_145      = VScreenColor(145, None, (0xaf,0xaf,0xaf))
            cls.term_146      = VScreenColor(146, None, (0xaf,0xaf,0xd7))
            cls.term_147      = VScreenColor(147, None, (0xaf,0xaf,0xff))
            cls.term_148      = VScreenColor(148, None, (0xaf,0xd7,0x00))
            cls.term_149      = VScreenColor(149, None, (0xaf,0xd7,0x5f))
            cls.term_150      = VScreenColor(150, None, (0xaf,0xd7,0x87))
            cls.term_151      = VScreenColor(151, None, (0xaf,0xd7,0xaf))
            cls.term_152      = VScreenColor(152, None, (0xaf,0xd7,0xd7))
            cls.term_153      = VScreenColor(153, None, (0xaf,0xd7,0xff))
            cls.term_154      = VScreenColor(154, None, (0xaf,0xff,0x00))
            cls.term_155      = VScreenColor(155, None, (0xaf,0xff,0x5f))
            cls.term_156      = VScreenColor(156, None, (0xaf,0xff,0x87))
            cls.term_157      = VScreenColor(157, None, (0xaf,0xff,0xaf))
            cls.term_158      = VScreenColor(158, None, (0xaf,0xff,0xd7))
            cls.term_159      = VScreenColor(159, None, (0xaf,0xff,0xff))
            cls.term_160      = VScreenColor(160, None, (0xd7,0x00,0x00))
            cls.term_161      = VScreenColor(161, None, (0xd7,0x00,0x5f))
            cls.term_162      = VScreenColor(162, None, (0xd7,0x00,0x87))
            cls.term_163      = VScreenColor(163, None, (0xd7,0x00,0xaf))
            cls.term_164      = VScreenColor(164, None, (0xd7,0x00,0xd7))
            cls.term_165      = VScreenColor(165, None, (0xd7,0x00,0xff))
            cls.term_166      = VScreenColor(166, None, (0xd7,0x5f,0x00))
            cls.term_167      = VScreenColor(167, None, (0xd7,0x5f,0x5f))
            cls.term_168      = VScreenColor(168, None, (0xd7,0x5f,0x87))
            cls.term_169      = VScreenColor(169, None, (0xd7,0x5f,0xaf))
            cls.term_170      = VScreenColor(170, None, (0xd7,0x5f,0xd7))
            cls.term_171      = VScreenColor(171, None, (0xd7,0x5f,0xff))
            cls.term_172      = VScreenColor(172, None, (0xd7,0x87,0x00))
            cls.term_173      = VScreenColor(173, None, (0xd7,0x87,0x5f))
            cls.term_174      = VScreenColor(174, None, (0xd7,0x87,0x87))
            cls.term_175      = VScreenColor(175, None, (0xd7,0x87,0xaf))
            cls.term_176      = VScreenColor(176, None, (0xd7,0x87,0xd7))
            cls.term_177      = VScreenColor(177, None, (0xd7,0x87,0xff))
            cls.term_178      = VScreenColor(178, None, (0xd7,0xaf,0x00))
            cls.term_179      = VScreenColor(179, None, (0xd7,0xaf,0x5f))
            cls.term_180      = VScreenColor(180, None, (0xd7,0xaf,0x87))
            cls.term_181      = VScreenColor(181, None, (0xd7,0xaf,0xaf))
            cls.term_182      = VScreenColor(182, None, (0xd7,0xaf,0xd7))
            cls.term_183      = VScreenColor(183, None, (0xd7,0xaf,0xff))
            cls.term_184      = VScreenColor(184, None, (0xd7,0xd7,0x00))
            cls.term_185      = VScreenColor(185, None, (0xd7,0xd7,0x5f))
            cls.term_186      = VScreenColor(186, None, (0xd7,0xd7,0x87))
            cls.term_187      = VScreenColor(187, None, (0xd7,0xd7,0xaf))
            cls.term_188      = VScreenColor(188, None, (0xd7,0xd7,0xd7))
            cls.term_189      = VScreenColor(189, None, (0xd7,0xd7,0xff))
            cls.term_190      = VScreenColor(190, None, (0xd7,0xff,0x00))
            cls.term_191      = VScreenColor(191, None, (0xd7,0xff,0x5f))
            cls.term_192      = VScreenColor(192, None, (0xd7,0xff,0x87))
            cls.term_193      = VScreenColor(193, None, (0xd7,0xff,0xaf))
            cls.term_194      = VScreenColor(194, None, (0xd7,0xff,0xd7))
            cls.term_195      = VScreenColor(195, None, (0xd7,0xff,0xff))
            cls.term_196      = VScreenColor(196, None, (0xff,0x00,0x00))
            cls.term_197      = VScreenColor(197, None, (0xff,0x00,0x5f))
            cls.term_198      = VScreenColor(198, None, (0xff,0x00,0x87))
            cls.term_199      = VScreenColor(199, None, (0xff,0x00,0xaf))
            cls.term_200      = VScreenColor(200, None, (0xff,0x00,0xd7))
            cls.term_201      = VScreenColor(201, None, (0xff,0x00,0xff))
            cls.term_202      = VScreenColor(202, None, (0xff,0x5f,0x00))
            cls.term_203      = VScreenColor(203, None, (0xff,0x5f,0x5f))
            cls.term_204      = VScreenColor(204, None, (0xff,0x5f,0x87))
            cls.term_205      = VScreenColor(205, None, (0xff,0x5f,0xaf))
            cls.term_206      = VScreenColor(206, None, (0xff,0x5f,0xd7))
            cls.term_207      = VScreenColor(207, None, (0xff,0x5f,0xff))
            cls.term_208      = VScreenColor(208, None, (0xff,0x87,0x00))
            cls.term_209      = VScreenColor(209, None, (0xff,0x87,0x5f))
            cls.term_210      = VScreenColor(210, None, (0xff,0x87,0x87))
            cls.term_211      = VScreenColor(211, None, (0xff,0x87,0xaf))
            cls.term_212      = VScreenColor(212, None, (0xff,0x87,0xd7))
            cls.term_213      = VScreenColor(213, None, (0xff,0x87,0xff))
            cls.term_214      = VScreenColor(214, None, (0xff,0xaf,0x00))
            cls.term_215      = VScreenColor(215, None, (0xff,0xaf,0x5f))
            cls.term_216      = VScreenColor(216, None, (0xff,0xaf,0x87))
            cls.term_217      = VScreenColor(217, None, (0xff,0xaf,0xaf))
            cls.term_218      = VScreenColor(218, None, (0xff,0xaf,0xd7))
            cls.term_219      = VScreenColor(219, None, (0xff,0xaf,0xff))
            cls.term_220      = VScreenColor(220, None, (0xff,0xd7,0x00))
            cls.term_221      = VScreenColor(221, None, (0xff,0xd7,0x5f))
            cls.term_222      = VScreenColor(222, None, (0xff,0xd7,0x87))
            cls.term_223      = VScreenColor(223, None, (0xff,0xd7,0xaf))
            cls.term_224      = VScreenColor(224, None, (0xff,0xd7,0xd7))
            cls.term_225      = VScreenColor(225, None, (0xff,0xd7,0xff))
            cls.term_226      = VScreenColor(226, None, (0xff,0xff,0x00))
            cls.term_227      = VScreenColor(227, None, (0xff,0xff,0x5f))
            cls.term_228      = VScreenColor(228, None, (0xff,0xff,0x87))
            cls.term_229      = VScreenColor(229, None, (0xff,0xff,0xaf))
            cls.term_230      = VScreenColor(230, None, (0xff,0xff,0xd7))
            cls.term_231      = VScreenColor(231, None, (0xff,0xff,0xff))
            cls.term_232      = VScreenColor(232, None, (0x08,0x08,0x08))
            cls.term_233      = VScreenColor(233, None, (0x12,0x12,0x12))
            cls.term_234      = VScreenColor(234, None, (0x1c,0x1c,0x1c))
            cls.term_235      = VScreenColor(235, None, (0x26,0x26,0x26))
            cls.term_236      = VScreenColor(236, None, (0x30,0x30,0x30))
            cls.term_237      = VScreenColor(237, None, (0x3a,0x3a,0x3a))
            cls.term_238      = VScreenColor(238, None, (0x44,0x44,0x44))
            cls.term_239      = VScreenColor(239, None, (0x4e,0x4e,0x4e))
            cls.term_240      = VScreenColor(240, None, (0x58,0x58,0x58))
            cls.term_241      = VScreenColor(241, None, (0x60,0x60,0x60))
            cls.term_242      = VScreenColor(242, None, (0x66,0x66,0x66))
            cls.term_243      = VScreenColor(243, None, (0x76,0x76,0x76))
            cls.term_244      = VScreenColor(244, None, (0x80,0x80,0x80))
            cls.term_245      = VScreenColor(245, None, (0x8a,0x8a,0x8a))
            cls.term_246      = VScreenColor(246, None, (0x94,0x94,0x94))
            cls.term_247      = VScreenColor(247, None, (0x9e,0x9e,0x9e))
            cls.term_248      = VScreenColor(248, None, (0xa8,0xa8,0xa8))
            cls.term_249      = VScreenColor(249, None, (0xb2,0xb2,0xb2))
            cls.term_250      = VScreenColor(250, None, (0xbc,0xbc,0xbc))
            cls.term_251      = VScreenColor(251, None, (0xc6,0xc6,0xc6))
            cls.term_252      = VScreenColor(252, None, (0xd0,0xd0,0xd0))
            cls.term_253      = VScreenColor(253, None, (0xda,0xda,0xda))
            cls.term_254      = VScreenColor(254, None, (0xe4,0xe4,0xe4))
            cls.term_255      = VScreenColor(255, None, (0xee,0xee,0xee))

            cls.pink = cls.term_210

    @classmethod
    def allColors(cls):
        return [c for c in list(cls.__dict__.values()) if isinstance(c, VScreenColor)]

class VScreenArea(object):
    def __init__(self, screen, rect):
        self._screen = screen
        self._rect = rect
        self.logger = logging.getLogger(self.__class__.__name__)
        if hasattr(self, "debug"):
            self.logger.setLevel(self.debug)
        else:
            self.logger.setLevel(logging.CRITICAL+1)

    def write(self, pos, string, fg_color=None, bg_color=None):
        rel_x, rel_y = pos
        w, h = self.size()

        if rel_y < 0 or rel_y >= h or rel_x >= w:
            self.logger.error("Out of bound in VScreenArea.write: pos=%s size=%s len=%d '%s'" % (str(pos), str(self.size()), len(string), string))
            return

        out_string = string
        if rel_x < 0:
            self.logger.error("Out of bound in VScreenArea.write: pos=%s size=%s len=%d '%s'" % (str(pos), str(self.size()), len(string), string))
            out_string = string[-rel_x:]
            rel_x = 0

        if len(out_string) == 0:
            return

        if (rel_x+len(out_string) > w):
            self.logger.error("Out of bound in VScreenArea.write: pos=%s size=%s len=%d '%s'" % (str(pos), str(self.size()), len(string), string))
            out_string = out_string[:w-rel_x]

        top_left_x, top_left_y = self.topLeft()
        self._screen.write( (rel_x+top_left_x, rel_y+top_left_y),
                             out_string,
                             fg_color,
                             bg_color)

    def setColors(self, pos, colors):
        rel_x, rel_y = pos
        w, h = self.size()

        if rel_y < 0 or rel_y >= h or rel_x >= w:
            self.logger.error("Out of bound in VScreenArea.setColors: pos=%s size=%s len=%d" % (str(pos), str(self.size()), len(colors)))
            return

        out_colors = colors
        if rel_x < 0:
            self.logger.error("Out of bound in VScreenArea.setColors: pos=%s size=%s len=%d" % (str(pos), str(self.size()), len(colors)))
            out_colors = colors[-rel_x:]
            rel_x = 0

        if len(out_colors) == 0:
            return

        if (rel_x+len(out_colors) > w):
            self.logger.error("Out of bound in VScreenArea.setColors: pos=%s size=%s len=%d" % (str(pos), str(self.size()), len(colors)))
            out_colors = out_colors[:w-rel_x]

        top_left_x, top_left_y = self.topLeft()
        self._screen.setColors( (rel_x+top_left_x, rel_y+top_left_y), out_colors)

    def rect(self):
        return self._rect

    def size(self):
        return (self._rect[Index.RECT_WIDTH], self._rect[Index.RECT_HEIGHT])

    def topLeft(self):
        return (self._rect[Index.RECT_X], self._rect[Index.RECT_Y])

    def width(self):
        return self._rect[Index.RECT_WIDTH]

    def height(self):
        return self._rect[Index.RECT_HEIGHT]

    def screen(self):
        return self._screen

    def erase(self):
        for y in range(self.height()):
            self.write( (0, y), ' '*self.width())

    def outOfBounds(self, pos):
        x, y = pos
        return (x >= self.size()[Index.SIZE_WIDTH] or y >= self.size()[Index.SIZE_HEIGHT] or x < 0 or y < 0)




class VHLayout(object):
    def __init__(self):
        self._widgets = []
        self._parent = None

    def addWidget(self, widget):
        self._widgets.append(widget)

    def apply(self):
        size = self.parent().size()
        available_size = size[0]/len(self._widgets)
        for i,w in enumerate(self._widgets):
            w.move(available_size*i,0)
            w.resize(available_size, size[1])

    def setParent(self, parent):
        self._parent = parent

    def parent(self):
        return self._parent

class VVLayout(object):
    def __init__(self):
        self._widgets = []
        self._parent = None

    def addWidget(self, widget):
        self._widgets.append(widget)

    def apply(self):
        size = self.parent().size()
        available_size = size[1]/len(self._widgets)
        remainder = size[1] % len(self._widgets)
        plot_pos = 0
        for i,w in enumerate(self._widgets):
            w.move(0, plot_pos)
            if remainder > 0:
                w.resize(size[0], available_size+1)
                remainder -= 1
                plot_pos += available_size + 1
            else:
                w.resize(size[0], available_size)
                plot_pos += available_size

    def setParent(self, parent):
        self._parent = parent

    def parent(self):
        return self._parent




def log(message):
    """
    Writes a message out in a log file.
    Use this to do printing while debugging, as you can't print while ncurses takes over the screen.
    """
    with open("vai.log", "a") as f:
        f.write(str(message)+'\n')




from vaitk import gui, core

class InfoHoverBox(core.VObject):
    """
    View class that observes the buffer. If the cursor hovers a line where the
    Linting has information, shows a popup containing the associated message
    """

    def __init__(self):
        self._buffer = None

        # The line for which the popup is visible
        self._current_shown_line = None
        self._hide_timer = core.VTimer()
        self._hide_timer.setInterval(3000)
        self._hide_timer.setSingleShot(True)
        self._hide_timer.timeout.connect(self._toolTipTimeOut)

    @property
    def buffer(self):
        return self._buffer

    @buffer.setter
    def buffer(self, buf):
        if buf is None:
            raise Exception("Buffer cannot be None")

        if self._hide_timer.isRunning():
            self._hide_timer.stop()

        gui.VToolTip.hideText()

        if self._buffer is not None:
            self._buffer.cursor.positionChanged.disconnect(self._cursorPositionChanged)

        self._buffer = buf
        self._buffer.cursor.positionChanged.connect(self._cursorPositionChanged)

    def _cursorPositionChanged(self, *args):
        if self._buffer is None:
            return

        cursor = self._buffer.cursor

        # If the cursor changed but remained on the same line, let things be
        if self._current_shown_line == cursor.pos[0]:
            return

        # Line changed. Stop the timer and hide the tooltip.
        self._hide_timer.stop()
        gui.VToolTip.hideText()

        # If there's information, show the new one
        lint = self._buffer.document.lineMetaInfo("LinterResult").data(cursor.pos[0])

        if lint is not None:
            pos_at_top = self._buffer.edit_area_model.document_pos_at_top
            self._current_shown_line = cursor.pos[0]
            gui.VToolTip.showText((0, cursor.pos[0]-pos_at_top[0]+1), " "+lint.message+" ")
            self._hide_timer.start()
        else:
            self._current_shown_line = None

    def _toolTipTimeOut(self):
        """
        Callback to hide the tooltip after a few seconds
        """
        gui.VToolTip.hideText()
        # Don't nullify the current show line, so that we don't show again
        # when moving along the line






import vaitk
from vaitk import gui, utils
from . import controllers
from .models.TextDocument import CharMeta
from . import models
from . import Search


class EditArea(gui.VWidget):
    def __init__(self, global_state, editor_controller, parent):
        super().__init__(parent)
        self._buffer = None

        self._controller = controllers.EditAreaController(self, global_state, editor_controller)
        self._color_schema = models.SyntaxColors(
                                    models.Configuration.get("colors.syntax_schema"),
                                    gui.VApplication.vApp.screen().numColors()
                             ).colorMap()

        self._visual_cursor_pos = (0,0)
        self.setFocusPolicy(vaitk.FocusPolicy.StrongFocus)

        self._icons = models.Icons.getCollection(
                                models.Configuration.get("icons.collection")
                                )

    # properties
    @property
    def buffer(self):
        return self._buffer

    @buffer.setter
    def buffer(self, buffer):
        if buffer is None:
            raise Exception("Cannot set buffer to None")

        self._buffer = buffer
        self._controller.buffer = buffer
        self.update()

    @property
    def visual_cursor_pos(self):
        return self._visual_cursor_pos

    @visual_cursor_pos.setter
    def visual_cursor_pos(self, cursor_pos):
        pos_x = utils.clamp(cursor_pos[0], 0, self.width()-1)
        pos_y = utils.clamp(cursor_pos[1], 0, self.height()-1)
        self._visual_cursor_pos = (pos_x, pos_y)
        gui.VCursor.setPos(self.mapToGlobal((pos_x, pos_y)))

    def paintEvent(self, event):
        painter = gui.VPainter(self)
        painter.erase()

        buffer = self._buffer
        if buffer is None:
            return

        w, h = self.size()
        pos_at_top = buffer.edit_area_model.document_pos_at_top
        visible_line_interval = (pos_at_top[0], pos_at_top[0]+h)
        cursor_pos = buffer.cursor.pos
        document = buffer.document

        # Find the current hovered word to set highlighting
        current_word, current_word_pos = document.wordAt(cursor_pos)
        word_entries = []
        if current_word_pos is not None:
            # find all the words only in the visible area
            word_entries = Search.findAll(document,
                                          current_word,
                                          line_interval=visible_line_interval,
                                          word=True)


        for visual_line_num, doc_line_num in enumerate(range(*visible_line_interval)):
            if doc_line_num > document.numLines():
                continue

            # Get the relevant text
            line_text = document.lineText(doc_line_num)[pos_at_top[1]-1:]
            line_text.replace('\n', ' ')

            # Apply colors. First through the Lexer designation
            colors = [(None, None, None)]*len(line_text)

            # Add markers for the indentation
            indent_spaces = len(line_text)-len(line_text.lstrip())
            for i in range(5, indent_spaces, 4):
                line_text = line_text[:i-1]+self._icons["EditArea.tab"]+line_text[i:]

            char_meta = document.charMeta( (doc_line_num,1))
            if CharMeta.LexerToken in char_meta:
                colors = [self._color_schema[tok] for tok in char_meta.get(CharMeta.LexerToken)]

            for i in range(5, indent_spaces, 4):
                colors[i-1] = (gui.VGlobalColor.term_303030, None, None)

            # Then, if there's a word, replace (None, None) entries with the highlight color
            word_entries_for_line = [x[1] for x in word_entries if x[0] == doc_line_num]
            for word_start in word_entries_for_line:
                for pos in range(word_start-1, word_start-1+len(current_word)):
                    if colors[pos] == (None, None, None):
                        colors[pos] = (gui.VGlobalColor.lightred, None, None)

            painter.drawText( (0, visual_line_num), line_text.replace('\n', ' '))
            painter.recolor((0, visual_line_num), colors[pos_at_top[1]-1:])

        self.visual_cursor_pos = (cursor_pos[1]-pos_at_top[1], cursor_pos[0]-pos_at_top[0])

    def keyEvent(self, event):
        self._controller.handleKeyEvent(event)

    def focusInEvent(self, event):
        gui.VCursor.setPos(self.mapToGlobal((self._visual_cursor_pos[0], self._visual_cursor_pos[1])))




import copy
import os
import json
import locale
from .. import paths

class Configuration:
    DEFAULTS = {
                 "colors.syntax_schema"     : "default",
                 "colors.status_bar.fg"     : "yellow",
                 "colors.status_bar.bg"     : "blue",
                 "colors.side_ruler.fg"     : "cyan",
                 "colors.side_ruler.bg"     : "transparent",
                 "icons.collection"         : "unicode1",
                 }

    # Singleton
    _instance = None

    # The file where to read the data
    _filename = None

    flags = { "has_wide_ncurses" : True }

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def setFilename(cls, filename):
        if cls._instance is not None and self._filename is not None:
            raise Exception("Configuration already initialized")

        cls._filename = filename

    @classmethod
    def filename(cls):
        if cls._filename is None:
            cls._filename = paths.configFile()
        return cls._filename

    @classmethod
    def save(cls):
        config = cls.instance()
        with open(cls._filename, "w") as f:
            f.write(json.dumps(config._config_dict, indent=4, sort_keys=True))

    def __init__(self):
        cls = self.__class__
        filename = cls.filename()

        try:
            with open(filename, "r") as f:
                merge_data = json.loads(f.read())
        except:
            merge_data = {}

        self._config_dict = copy.deepcopy(Configuration.DEFAULTS)
        self._config_dict.update(merge_data)

    def __getitem__(self, key):
        return self._config_dict[key]

    @classmethod
    def get(cls, key):
        # We override, but not alter, the configuration if the current encoding
        # is not supporting utf-8 encoding, or if we don't have wide ncurses
        if (key == "icons.collection" and locale.getpreferredencoding(False) != "UTF-8") \
            or not cls.flags.get("has_wide_ncurses"):
            return "ascii"
        return cls.instance()[key]






class EditorMode:
    """
    Separate enum class to hold the current editor mode
    """
    COMMAND = 0         # Normal, command mode, waiting for something to happen
    COMMAND_INPUT = 1   # When : has been pressed
    INSERT = 2          # When in insert mode you can type in the document.
    REPLACE = 3         # Single place replacement for r<char>
    VISUAL_BLOCK = 4    # For visual block mode (selection). Not yet used.
    VISUAL_LINE = 5     # For visual line mode (selection). Not yet used.
    VISUAL = 6          # For visual char mode (selection). Not yet used.
    DELETE = 7          # When d has been pressed and waiting for specification on what to delete
    SEARCH_FORWARD = 8  # When / has been pressed
    SEARCH_BACKWARD = 9 # When ? has been pressed
    GO = 10             # When g is pressed and it's waiting for the second g to go at BOF
    YANK = 11           # When y has been pressed and waiting for the second y.
    ZETA = 12           # When capital Z has been pressed and waiting for the second Z
    BOOKMARK = 13       #
    GOTOBOOKMARK = 14   #




from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class InsertFileCommand(BufferCommand):
    def __init__(self, buffer, filename):
        super().__init__(buffer)
        self._filename = filename
        self._how_many = None

    def execute(self):
        document = self._document
        cursor = self._cursor
        line_pos = cursor.line

        try:
            with open(self._filename,'r') as f:
                lines = f.readlines()
        except:
            return CommandResult(False, None)

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()
        cursor.toPos(pos)

        self._how_many = len(lines)
        document.insertLines(line_pos, lines)
        document.lineMetaInfo("Change").setData(["added"] * self._how_many, line_pos)

        return CommandResult(True, None)

    def undo(self):
        self._document.deleteLines(self.savedCursorPos()[0]+1, self._how_many)
        super().undo()





class VColor:
    def __init__(self, rgb):
        self._rgb = rgb

    @property
    def rgb(self):
        return self._rgb

    def hexString(self):
        return "%0.2X%0.2X%0.2X" % self.rgb

    @property
    def r(self):
        return self._rgb[0]

    @property
    def g(self):
        return self._rgb[1]

    @property
    def b(self):
        return self._rgb[2]

    @staticmethod
    def distance(color1, color2):
        return (color1.r - color2.r)**2 + (color1.g - color2.g)**2 + (color1.b - color2.b)**2

    class tuple:
        @staticmethod
        def distance(color1, color2):
            return (color1[0] - color2[0])**2 + (color1[1] - color2[1])**2 + (color1[2] - color2[2])**2

class VGlobalColor:
    """This class defines colors according to the rgb palette"""
    term_0        = VColor(rgb=(0x00,0x00,0x00))
    term_1        = VColor(rgb=(0x80,0x00,0x00))
    term_2        = VColor(rgb=(0x00,0x80,0x00))
    term_3        = VColor(rgb=(0x80,0x80,0x00))
    term_4        = VColor(rgb=(0x00,0x00,0x80))
    term_5        = VColor(rgb=(0x80,0x00,0x80))
    term_6        = VColor(rgb=(0x00,0x80,0x80))
    term_7        = VColor(rgb=(0xc0,0xc0,0xc0))
    term_8        = VColor(rgb=(0x80,0x80,0x80))
    term_9        = VColor(rgb=(0xff,0x00,0x00))
    term_10       = VColor(rgb=(0x00,0xff,0x00))
    term_11       = VColor(rgb=(0xff,0xff,0x00))
    term_12       = VColor(rgb=(0x00,0x00,0xff))
    term_13       = VColor(rgb=(0xff,0x00,0xff))
    term_14       = VColor(rgb=(0x00,0xff,0xff))
    term_15       = VColor(rgb=(0xff,0xff,0xff))
    term_16       = VColor(rgb=(0x00,0x00,0x00))
    term_17       = VColor(rgb=(0x00,0x00,0x5f))
    term_18       = VColor(rgb=(0x00,0x00,0x87))
    term_19       = VColor(rgb=(0x00,0x00,0xaf))
    term_20       = VColor(rgb=(0x00,0x00,0xd7))
    term_21       = VColor(rgb=(0x00,0x00,0xff))
    term_22       = VColor(rgb=(0x00,0x5f,0x00))
    term_23       = VColor(rgb=(0x00,0x5f,0x5f))
    term_24       = VColor(rgb=(0x00,0x5f,0x87))
    term_25       = VColor(rgb=(0x00,0x5f,0xaf))
    term_26       = VColor(rgb=(0x00,0x5f,0xd7))
    term_27       = VColor(rgb=(0x00,0x5f,0xff))
    term_28       = VColor(rgb=(0x00,0x87,0x00))
    term_29       = VColor(rgb=(0x00,0x87,0x5f))
    term_30       = VColor(rgb=(0x00,0x87,0x87))
    term_31       = VColor(rgb=(0x00,0x87,0xaf))
    term_32       = VColor(rgb=(0x00,0x87,0xd7))
    term_33       = VColor(rgb=(0x00,0x87,0xff))
    term_34       = VColor(rgb=(0x00,0xaf,0x00))
    term_35       = VColor(rgb=(0x00,0xaf,0x5f))
    term_36       = VColor(rgb=(0x00,0xaf,0x87))
    term_37       = VColor(rgb=(0x00,0xaf,0xaf))
    term_38       = VColor(rgb=(0x00,0xaf,0xd7))
    term_39       = VColor(rgb=(0x00,0xaf,0xff))
    term_40       = VColor(rgb=(0x00,0xd7,0x00))
    term_41       = VColor(rgb=(0x00,0xd7,0x5f))
    term_42       = VColor(rgb=(0x00,0xd7,0x87))
    term_43       = VColor(rgb=(0x00,0xd7,0xaf))
    term_44       = VColor(rgb=(0x00,0xd7,0xd7))
    term_45       = VColor(rgb=(0x00,0xd7,0xff))
    term_46       = VColor(rgb=(0x00,0xff,0x00))
    term_47       = VColor(rgb=(0x00,0xff,0x5f))
    term_48       = VColor(rgb=(0x00,0xff,0x87))
    term_49       = VColor(rgb=(0x00,0xff,0xaf))
    term_50       = VColor(rgb=(0x00,0xff,0xd7))
    term_51       = VColor(rgb=(0x00,0xff,0xff))
    term_52       = VColor(rgb=(0x5f,0x00,0x00))
    term_53       = VColor(rgb=(0x5f,0x00,0x5f))
    term_54       = VColor(rgb=(0x5f,0x00,0x87))
    term_55       = VColor(rgb=(0x5f,0x00,0xaf))
    term_56       = VColor(rgb=(0x5f,0x00,0xd7))
    term_57       = VColor(rgb=(0x5f,0x00,0xff))
    term_58       = VColor(rgb=(0x5f,0x5f,0x00))
    term_59       = VColor(rgb=(0x5f,0x5f,0x5f))
    term_60       = VColor(rgb=(0x5f,0x5f,0x87))
    term_61       = VColor(rgb=(0x5f,0x5f,0xaf))
    term_62       = VColor(rgb=(0x5f,0x5f,0xd7))
    term_63       = VColor(rgb=(0x5f,0x5f,0xff))
    term_64       = VColor(rgb=(0x5f,0x87,0x00))
    term_65       = VColor(rgb=(0x5f,0x87,0x5f))
    term_66       = VColor(rgb=(0x5f,0x87,0x87))
    term_67       = VColor(rgb=(0x5f,0x87,0xaf))
    term_68       = VColor(rgb=(0x5f,0x87,0xd7))
    term_69       = VColor(rgb=(0x5f,0x87,0xff))
    term_70       = VColor(rgb=(0x5f,0xaf,0x00))
    term_71       = VColor(rgb=(0x5f,0xaf,0x5f))
    term_72       = VColor(rgb=(0x5f,0xaf,0x87))
    term_73       = VColor(rgb=(0x5f,0xaf,0xaf))
    term_74       = VColor(rgb=(0x5f,0xaf,0xd7))
    term_75       = VColor(rgb=(0x5f,0xaf,0xff))
    term_76       = VColor(rgb=(0x5f,0xd7,0x00))
    term_77       = VColor(rgb=(0x5f,0xd7,0x5f))
    term_78       = VColor(rgb=(0x5f,0xd7,0x87))
    term_79       = VColor(rgb=(0x5f,0xd7,0xaf))
    term_80       = VColor(rgb=(0x5f,0xd7,0xd7))
    term_81       = VColor(rgb=(0x5f,0xd7,0xff))
    term_82       = VColor(rgb=(0x5f,0xff,0x00))
    term_83       = VColor(rgb=(0x5f,0xff,0x5f))
    term_84       = VColor(rgb=(0x5f,0xff,0x87))
    term_85       = VColor(rgb=(0x5f,0xff,0xaf))
    term_86       = VColor(rgb=(0x5f,0xff,0xd7))
    term_87       = VColor(rgb=(0x5f,0xff,0xff))
    term_88       = VColor(rgb=(0x87,0x00,0x00))
    term_89       = VColor(rgb=(0x87,0x00,0x5f))
    term_90       = VColor(rgb=(0x87,0x00,0x87))
    term_91       = VColor(rgb=(0x87,0x00,0xaf))
    term_92       = VColor(rgb=(0x87,0x00,0xd7))
    term_93       = VColor(rgb=(0x87,0x00,0xff))
    term_94       = VColor(rgb=(0x87,0x5f,0x00))
    term_95       = VColor(rgb=(0x87,0x5f,0x5f))
    term_96       = VColor(rgb=(0x87,0x5f,0x87))
    term_97       = VColor(rgb=(0x87,0x5f,0xaf))
    term_98       = VColor(rgb=(0x87,0x5f,0xd7))
    term_99       = VColor(rgb=(0x87,0x5f,0xff))
    term_100      = VColor(rgb=(0x87,0x87,0x00))
    term_101      = VColor(rgb=(0x87,0x87,0x5f))
    term_102      = VColor(rgb=(0x87,0x87,0x87))
    term_103      = VColor(rgb=(0x87,0x87,0xaf))
    term_104      = VColor(rgb=(0x87,0x87,0xd7))
    term_105      = VColor(rgb=(0x87,0x87,0xff))
    term_106      = VColor(rgb=(0x87,0xaf,0x00))
    term_107      = VColor(rgb=(0x87,0xaf,0x5f))
    term_108      = VColor(rgb=(0x87,0xaf,0x87))
    term_109      = VColor(rgb=(0x87,0xaf,0xaf))
    term_110      = VColor(rgb=(0x87,0xaf,0xd7))
    term_111      = VColor(rgb=(0x87,0xaf,0xff))
    term_112      = VColor(rgb=(0x87,0xd7,0x00))
    term_113      = VColor(rgb=(0x87,0xd7,0x5f))
    term_114      = VColor(rgb=(0x87,0xd7,0x87))
    term_115      = VColor(rgb=(0x87,0xd7,0xaf))
    term_116      = VColor(rgb=(0x87,0xd7,0xd7))
    term_117      = VColor(rgb=(0x87,0xd7,0xff))
    term_118      = VColor(rgb=(0x87,0xff,0x00))
    term_119      = VColor(rgb=(0x87,0xff,0x5f))
    term_120      = VColor(rgb=(0x87,0xff,0x87))
    term_121      = VColor(rgb=(0x87,0xff,0xaf))
    term_122      = VColor(rgb=(0x87,0xff,0xd7))
    term_123      = VColor(rgb=(0x87,0xff,0xff))
    term_124      = VColor(rgb=(0xaf,0x00,0x00))
    term_125      = VColor(rgb=(0xaf,0x00,0x5f))
    term_126      = VColor(rgb=(0xaf,0x00,0x87))
    term_127      = VColor(rgb=(0xaf,0x00,0xaf))
    term_128      = VColor(rgb=(0xaf,0x00,0xd7))
    term_129      = VColor(rgb=(0xaf,0x00,0xff))
    term_130      = VColor(rgb=(0xaf,0x5f,0x00))
    term_131      = VColor(rgb=(0xaf,0x5f,0x5f))
    term_132      = VColor(rgb=(0xaf,0x5f,0x87))
    term_133      = VColor(rgb=(0xaf,0x5f,0xaf))
    term_134      = VColor(rgb=(0xaf,0x5f,0xd7))
    term_135      = VColor(rgb=(0xaf,0x5f,0xff))
    term_136      = VColor(rgb=(0xaf,0x87,0x00))
    term_137      = VColor(rgb=(0xaf,0x87,0x5f))
    term_138      = VColor(rgb=(0xaf,0x87,0x87))
    term_139      = VColor(rgb=(0xaf,0x87,0xaf))
    term_140      = VColor(rgb=(0xaf,0x87,0xd7))
    term_141      = VColor(rgb=(0xaf,0x87,0xff))
    term_142      = VColor(rgb=(0xaf,0xaf,0x00))
    term_143      = VColor(rgb=(0xaf,0xaf,0x5f))
    term_144      = VColor(rgb=(0xaf,0xaf,0x87))
    term_145      = VColor(rgb=(0xaf,0xaf,0xaf))
    term_146      = VColor(rgb=(0xaf,0xaf,0xd7))
    term_147      = VColor(rgb=(0xaf,0xaf,0xff))
    term_148      = VColor(rgb=(0xaf,0xd7,0x00))
    term_149      = VColor(rgb=(0xaf,0xd7,0x5f))
    term_150      = VColor(rgb=(0xaf,0xd7,0x87))
    term_151      = VColor(rgb=(0xaf,0xd7,0xaf))
    term_152      = VColor(rgb=(0xaf,0xd7,0xd7))
    term_153      = VColor(rgb=(0xaf,0xd7,0xff))
    term_154      = VColor(rgb=(0xaf,0xff,0x00))
    term_155      = VColor(rgb=(0xaf,0xff,0x5f))
    term_156      = VColor(rgb=(0xaf,0xff,0x87))
    term_157      = VColor(rgb=(0xaf,0xff,0xaf))
    term_158      = VColor(rgb=(0xaf,0xff,0xd7))
    term_159      = VColor(rgb=(0xaf,0xff,0xff))
    term_160      = VColor(rgb=(0xd7,0x00,0x00))
    term_161      = VColor(rgb=(0xd7,0x00,0x5f))
    term_162      = VColor(rgb=(0xd7,0x00,0x87))
    term_163      = VColor(rgb=(0xd7,0x00,0xaf))
    term_164      = VColor(rgb=(0xd7,0x00,0xd7))
    term_165      = VColor(rgb=(0xd7,0x00,0xff))
    term_166      = VColor(rgb=(0xd7,0x5f,0x00))
    term_167      = VColor(rgb=(0xd7,0x5f,0x5f))
    term_168      = VColor(rgb=(0xd7,0x5f,0x87))
    term_169      = VColor(rgb=(0xd7,0x5f,0xaf))
    term_170      = VColor(rgb=(0xd7,0x5f,0xd7))
    term_171      = VColor(rgb=(0xd7,0x5f,0xff))
    term_172      = VColor(rgb=(0xd7,0x87,0x00))
    term_173      = VColor(rgb=(0xd7,0x87,0x5f))
    term_174      = VColor(rgb=(0xd7,0x87,0x87))
    term_175      = VColor(rgb=(0xd7,0x87,0xaf))
    term_176      = VColor(rgb=(0xd7,0x87,0xd7))
    term_177      = VColor(rgb=(0xd7,0x87,0xff))
    term_178      = VColor(rgb=(0xd7,0xaf,0x00))
    term_179      = VColor(rgb=(0xd7,0xaf,0x5f))
    term_180      = VColor(rgb=(0xd7,0xaf,0x87))
    term_181      = VColor(rgb=(0xd7,0xaf,0xaf))
    term_182      = VColor(rgb=(0xd7,0xaf,0xd7))
    term_183      = VColor(rgb=(0xd7,0xaf,0xff))
    term_184      = VColor(rgb=(0xd7,0xd7,0x00))
    term_185      = VColor(rgb=(0xd7,0xd7,0x5f))
    term_186      = VColor(rgb=(0xd7,0xd7,0x87))
    term_187      = VColor(rgb=(0xd7,0xd7,0xaf))
    term_188      = VColor(rgb=(0xd7,0xd7,0xd7))
    term_189      = VColor(rgb=(0xd7,0xd7,0xff))
    term_190      = VColor(rgb=(0xd7,0xff,0x00))
    term_191      = VColor(rgb=(0xd7,0xff,0x5f))
    term_192      = VColor(rgb=(0xd7,0xff,0x87))
    term_193      = VColor(rgb=(0xd7,0xff,0xaf))
    term_194      = VColor(rgb=(0xd7,0xff,0xd7))
    term_195      = VColor(rgb=(0xd7,0xff,0xff))
    term_196      = VColor(rgb=(0xff,0x00,0x00))
    term_197      = VColor(rgb=(0xff,0x00,0x5f))
    term_198      = VColor(rgb=(0xff,0x00,0x87))
    term_199      = VColor(rgb=(0xff,0x00,0xaf))
    term_200      = VColor(rgb=(0xff,0x00,0xd7))
    term_201      = VColor(rgb=(0xff,0x00,0xff))
    term_202      = VColor(rgb=(0xff,0x5f,0x00))
    term_203      = VColor(rgb=(0xff,0x5f,0x5f))
    term_204      = VColor(rgb=(0xff,0x5f,0x87))
    term_205      = VColor(rgb=(0xff,0x5f,0xaf))
    term_206      = VColor(rgb=(0xff,0x5f,0xd7))
    term_207      = VColor(rgb=(0xff,0x5f,0xff))
    term_208      = VColor(rgb=(0xff,0x87,0x00))
    term_209      = VColor(rgb=(0xff,0x87,0x5f))
    term_210      = VColor(rgb=(0xff,0x87,0x87))
    term_211      = VColor(rgb=(0xff,0x87,0xaf))
    term_212      = VColor(rgb=(0xff,0x87,0xd7))
    term_213      = VColor(rgb=(0xff,0x87,0xff))
    term_214      = VColor(rgb=(0xff,0xaf,0x00))
    term_215      = VColor(rgb=(0xff,0xaf,0x5f))
    term_216      = VColor(rgb=(0xff,0xaf,0x87))
    term_217      = VColor(rgb=(0xff,0xaf,0xaf))
    term_218      = VColor(rgb=(0xff,0xaf,0xd7))
    term_219      = VColor(rgb=(0xff,0xaf,0xff))
    term_220      = VColor(rgb=(0xff,0xd7,0x00))
    term_221      = VColor(rgb=(0xff,0xd7,0x5f))
    term_222      = VColor(rgb=(0xff,0xd7,0x87))
    term_223      = VColor(rgb=(0xff,0xd7,0xaf))
    term_224      = VColor(rgb=(0xff,0xd7,0xd7))
    term_225      = VColor(rgb=(0xff,0xd7,0xff))
    term_226      = VColor(rgb=(0xff,0xff,0x00))
    term_227      = VColor(rgb=(0xff,0xff,0x5f))
    term_228      = VColor(rgb=(0xff,0xff,0x87))
    term_229      = VColor(rgb=(0xff,0xff,0xaf))
    term_230      = VColor(rgb=(0xff,0xff,0xd7))
    term_231      = VColor(rgb=(0xff,0xff,0xff))
    term_232      = VColor(rgb=(0x08,0x08,0x08))
    term_233      = VColor(rgb=(0x12,0x12,0x12))
    term_234      = VColor(rgb=(0x1c,0x1c,0x1c))
    term_235      = VColor(rgb=(0x26,0x26,0x26))
    term_236      = VColor(rgb=(0x30,0x30,0x30))
    term_237      = VColor(rgb=(0x3a,0x3a,0x3a))
    term_238      = VColor(rgb=(0x44,0x44,0x44))
    term_239      = VColor(rgb=(0x4e,0x4e,0x4e))
    term_240      = VColor(rgb=(0x58,0x58,0x58))
    term_241      = VColor(rgb=(0x60,0x60,0x60))
    term_242      = VColor(rgb=(0x66,0x66,0x66))
    term_243      = VColor(rgb=(0x76,0x76,0x76))
    term_244      = VColor(rgb=(0x80,0x80,0x80))
    term_245      = VColor(rgb=(0x8a,0x8a,0x8a))
    term_246      = VColor(rgb=(0x94,0x94,0x94))
    term_247      = VColor(rgb=(0x9e,0x9e,0x9e))
    term_248      = VColor(rgb=(0xa8,0xa8,0xa8))
    term_249      = VColor(rgb=(0xb2,0xb2,0xb2))
    term_250      = VColor(rgb=(0xbc,0xbc,0xbc))
    term_251      = VColor(rgb=(0xc6,0xc6,0xc6))
    term_252      = VColor(rgb=(0xd0,0xd0,0xd0))
    term_253      = VColor(rgb=(0xda,0xda,0xda))
    term_254      = VColor(rgb=(0xe4,0xe4,0xe4))
    term_255      = VColor(rgb=(0xee,0xee,0xee))

    term_000000 = term_0        
    term_800000 = term_1        
    term_008000 = term_2        
    term_808000 = term_3        
    term_000080 = term_4        
    term_800080 = term_5        
    term_008080 = term_6        
    term_c0c0c0 = term_7        
    term_808080 = term_8        
    term_ff0000 = term_9        
    term_00ff00 = term_10       
    term_ffff00 = term_11       
    term_0000ff = term_12       
    term_ff00ff = term_13       
    term_00ffff = term_14       
    term_ffffff = term_15       
    term_000000 = term_16       
    term_00005f = term_17       
    term_000087 = term_18       
    term_0000af = term_19       
    term_0000d7 = term_20       
    term_0000ff = term_21       
    term_005f00 = term_22       
    term_005f5f = term_23       
    term_005f87 = term_24       
    term_005faf = term_25       
    term_005fd7 = term_26       
    term_005fff = term_27       
    term_008700 = term_28       
    term_00875f = term_29       
    term_008787 = term_30       
    term_0087af = term_31       
    term_0087d7 = term_32       
    term_0087ff = term_33       
    term_00af00 = term_34       
    term_00af5f = term_35       
    term_00af87 = term_36       
    term_00afaf = term_37       
    term_00afd7 = term_38       
    term_00afff = term_39       
    term_00d700 = term_40       
    term_00d75f = term_41       
    term_00d787 = term_42       
    term_00d7af = term_43       
    term_00d7d7 = term_44       
    term_00d7ff = term_45       
    term_00ff00 = term_46       
    term_00ff5f = term_47       
    term_00ff87 = term_48       
    term_00ffaf = term_49       
    term_00ffd7 = term_50       
    term_00ffff = term_51       
    term_5f0000 = term_52       
    term_5f005f = term_53       
    term_5f0087 = term_54       
    term_5f00af = term_55       
    term_5f00d7 = term_56       
    term_5f00ff = term_57       
    term_5f5f00 = term_58       
    term_5f5f5f = term_59       
    term_5f5f87 = term_60       
    term_5f5faf = term_61       
    term_5f5fd7 = term_62       
    term_5f5fff = term_63       
    term_5f8700 = term_64       
    term_5f875f = term_65       
    term_5f8787 = term_66       
    term_5f87af = term_67       
    term_5f87d7 = term_68       
    term_5f87ff = term_69       
    term_5faf00 = term_70       
    term_5faf5f = term_71       
    term_5faf87 = term_72       
    term_5fafaf = term_73       
    term_5fafd7 = term_74       
    term_5fafff = term_75       
    term_5fd700 = term_76       
    term_5fd75f = term_77       
    term_5fd787 = term_78       
    term_5fd7af = term_79       
    term_5fd7d7 = term_80       
    term_5fd7ff = term_81       
    term_5fff00 = term_82       
    term_5fff5f = term_83       
    term_5fff87 = term_84       
    term_5fffaf = term_85       
    term_5fffd7 = term_86       
    term_5fffff = term_87       
    term_870000 = term_88       
    term_87005f = term_89       
    term_870087 = term_90       
    term_8700af = term_91       
    term_8700d7 = term_92       
    term_8700ff = term_93       
    term_875f00 = term_94       
    term_875f5f = term_95       
    term_875f87 = term_96       
    term_875faf = term_97       
    term_875fd7 = term_98       
    term_875fff = term_99       
    term_878700 = term_100      
    term_87875f = term_101      
    term_878787 = term_102      
    term_8787af = term_103      
    term_8787d7 = term_104      
    term_8787ff = term_105      
    term_87af00 = term_106      
    term_87af5f = term_107      
    term_87af87 = term_108      
    term_87afaf = term_109      
    term_87afd7 = term_110      
    term_87afff = term_111      
    term_87d700 = term_112      
    term_87d75f = term_113      
    term_87d787 = term_114      
    term_87d7af = term_115      
    term_87d7d7 = term_116      
    term_87d7ff = term_117      
    term_87ff00 = term_118      
    term_87ff5f = term_119      
    term_87ff87 = term_120      
    term_87ffaf = term_121      
    term_87ffd7 = term_122      
    term_87ffff = term_123      
    term_af0000 = term_124      
    term_af005f = term_125      
    term_af0087 = term_126      
    term_af00af = term_127      
    term_af00d7 = term_128      
    term_af00ff = term_129      
    term_af5f00 = term_130      
    term_af5f5f = term_131      
    term_af5f87 = term_132      
    term_af5faf = term_133      
    term_af5fd7 = term_134      
    term_af5fff = term_135      
    term_af8700 = term_136      
    term_af875f = term_137      
    term_af8787 = term_138      
    term_af87af = term_139      
    term_af87d7 = term_140      
    term_af87ff = term_141      
    term_afaf00 = term_142      
    term_afaf5f = term_143      
    term_afaf87 = term_144      
    term_afafaf = term_145      
    term_afafd7 = term_146      
    term_afafff = term_147      
    term_afd700 = term_148      
    term_afd75f = term_149      
    term_afd787 = term_150      
    term_afd7af = term_151      
    term_afd7d7 = term_152      
    term_afd7ff = term_153      
    term_afff00 = term_154      
    term_afff5f = term_155      
    term_afff87 = term_156      
    term_afffaf = term_157      
    term_afffd7 = term_158      
    term_afffff = term_159      
    term_d70000 = term_160      
    term_d7005f = term_161      
    term_d70087 = term_162      
    term_d700af = term_163      
    term_d700d7 = term_164      
    term_d700ff = term_165      
    term_d75f00 = term_166      
    term_d75f5f = term_167      
    term_d75f87 = term_168      
    term_d75faf = term_169      
    term_d75fd7 = term_170      
    term_d75fff = term_171      
    term_d78700 = term_172      
    term_d7875f = term_173      
    term_d78787 = term_174      
    term_d787af = term_175      
    term_d787d7 = term_176      
    term_d787ff = term_177      
    term_d7af00 = term_178      
    term_d7af5f = term_179      
    term_d7af87 = term_180      
    term_d7afaf = term_181      
    term_d7afd7 = term_182      
    term_d7afff = term_183      
    term_d7d700 = term_184      
    term_d7d75f = term_185      
    term_d7d787 = term_186      
    term_d7d7af = term_187      
    term_d7d7d7 = term_188      
    term_d7d7ff = term_189      
    term_d7ff00 = term_190      
    term_d7ff5f = term_191      
    term_d7ff87 = term_192      
    term_d7ffaf = term_193      
    term_d7ffd7 = term_194      
    term_d7ffff = term_195      
    term_ff0000 = term_196      
    term_ff005f = term_197      
    term_ff0087 = term_198      
    term_ff00af = term_199      
    term_ff00d7 = term_200      
    term_ff00ff = term_201      
    term_ff5f00 = term_202      
    term_ff5f5f = term_203      
    term_ff5f87 = term_204      
    term_ff5faf = term_205      
    term_ff5fd7 = term_206      
    term_ff5fff = term_207      
    term_ff8700 = term_208      
    term_ff875f = term_209      
    term_ff8787 = term_210      
    term_ff87af = term_211      
    term_ff87d7 = term_212      
    term_ff87ff = term_213      
    term_ffaf00 = term_214      
    term_ffaf5f = term_215      
    term_ffaf87 = term_216      
    term_ffafaf = term_217      
    term_ffafd7 = term_218      
    term_ffafff = term_219      
    term_ffd700 = term_220      
    term_ffd75f = term_221      
    term_ffd787 = term_222      
    term_ffd7af = term_223      
    term_ffd7d7 = term_224      
    term_ffd7ff = term_225      
    term_ffff00 = term_226      
    term_ffff5f = term_227      
    term_ffff87 = term_228      
    term_ffffaf = term_229      
    term_ffffd7 = term_230      
    term_ffffff = term_231      
    term_080808 = term_232      
    term_121212 = term_233      
    term_1c1c1c = term_234      
    term_262626 = term_235      
    term_303030 = term_236      
    term_3a3a3a = term_237      
    term_444444 = term_238      
    term_4e4e4e = term_239      
    term_585858 = term_240      
    term_606060 = term_241      
    term_666666 = term_242      
    term_767676 = term_243      
    term_808080 = term_244      
    term_8a8a8a = term_245      
    term_949494 = term_246      
    term_9e9e9e = term_247      
    term_a8a8a8 = term_248      
    term_b2b2b2 = term_249      
    term_bcbcbc = term_250      
    term_c6c6c6 = term_251      
    term_d0d0d0 = term_252      
    term_dadada = term_253      
    term_e4e4e4 = term_254      
    term_eeeeee = term_255      

    transparent = None

    black = term_0
    red = term_1
    green = term_2
    brown = term_3
    blue = term_4
    magenta = term_5 
    cyan = term_6
    white = term_7
    grey = term_8
    lightred = term_9
    lightgreen = term_10
    yellow = term_11
    lightblue = term_12
    lightmagenta = term_13
    lightcyan = term_14
    white = term_15

    pink = term_ff8787
    darkred = term_5f0000
    darkgreen = term_005f00
    darkblue = term_00005f
    darkmagenta = term_5f005f
    darkcyan = term_005f5f


    @staticmethod
    def nameToColor(name):
        """Perform lookup of the color by string. Returns None if the lookup fails."""
        return VGlobalColor.__dict__.get(name)




from vaitk import core
from .EditorMode import EditorMode

class GlobalState(core.VObject):
    """
    Represents global state of the editor that is not dependent on
    the currently selected buffer.
    """
    def __init__(self):
        super().__init__()
        self._editor_mode = EditorMode.COMMAND
        self._current_search = None
        self._clipboard = None

        self.editorModeChanged = core.VSignal(self)

    @property
    def editor_mode(self):
        return self._editor_mode

    @editor_mode.setter
    def editor_mode(self, mode):
        if self._editor_mode != mode:
            self._editor_mode = mode
            self.editorModeChanged.emit(mode)

    @property
    def current_search(self):
        return self._current_search

    @current_search.setter
    def current_search(self, search):
        assert(len(search) == 2)
        self._current_search = search

    @property
    def clipboard(self):
        return self._clipboard

    @clipboard.setter
    def clipboard(self, text):
        self._clipboard = text




import os
import curses

class FocusPolicy:
    NoFocus = 0
    StrongFocus = 11

class Orientation:
    Horizontal = 1
    Vertical = 2

class LineStyle:
    NoLine = 0
    Full   = 1

class CornerCapStyle:
    NoCap = 0
    Plus = 1

class LineCapStyle:
    NoCap = 0
    Plus  = 1

class Alignment:
    AlignLeft    =  0x1
    AlignRight   =  0x2
    AlignHCenter =  0x4
    AlignTop     = 0x20
    AlignBottom  = 0x40
    AlignVCenter = 0x80
    AlignCenter  = AlignHCenter|AlignVCenter

class Key:
    Key_Escape    = 0x01000000
    Key_Tab       = 0x01000001
    Key_Backtab   = 0x01000002
    Key_Backspace = 0x01000003
    Key_Return    = 0x01000004
    Key_Enter     = 0x01000005
    Key_Insert    = 0x01000006
    Key_Delete    = 0x01000007
    Key_Pause     = 0x01000008
    Key_Print     = 0x01000009
    Key_SysReq    = 0x0100000a
    Key_Clear     = 0x0100000b
    Key_Home      = 0x01000010
    Key_End       = 0x01000011
    Key_Left      = 0x01000012
    Key_Up        =  0x01000013
    Key_Right     =  0x01000014
    Key_Down      =  0x01000015
    Key_PageUp    =  0x01000016
    Key_PageDown  =  0x01000017
    Key_Shift     =  0x01000020
    Key_Control   =0x01000021
    Key_Meta      =0x01000022
    Key_Alt       =0x01000023
    Key_AltGr     =0x01001103
    Key_CapsLock  =   0x01000024
    Key_NumLock   =0x01000025
    Key_ScrollLock=  0x01000026
    Key_F1        =0x01000030
    Key_F2        =0x01000031
    Key_F3        = 0x01000032
    Key_F4        = 0x01000033
    Key_F5        = 0x01000034
    Key_F6        = 0x01000035
    Key_F7        = 0x01000036
    Key_F8        = 0x01000037
    Key_F9        = 0x01000038
    Key_F10       = 0x01000039
    Key_F11       = 0x0100003a
    Key_F12       = 0x0100003b
    Key_F13       = 0x0100003c
    Key_F14       = 0x0100003d
    Key_F15       = 0x0100003e
    Key_F16       = 0x0100003f
    Key_F17       = 0x01000040
    Key_F18       = 0x01000041
    Key_F19       = 0x01000042
    Key_F20       = 0x01000043
    Key_F21       = 0x01000044
    Key_F22       = 0x01000045
    Key_F23       = 0x01000046
    Key_F24       = 0x01000047
    Key_F25       = 0x01000048
    Key_F26       = 0x01000049
    Key_F27       = 0x0100004a
    Key_F28       = 0x0100004b
    Key_F29       = 0x0100004c
    Key_F30       = 0x0100004d
    Key_F31       = 0x0100004e
    Key_F32       = 0x0100004f
    Key_F33       = 0x01000050
    Key_F34       = 0x01000051
    Key_F35       = 0x01000052
    Key_Super_L   =0x01000053
    Key_Super_R   =0x01000054
    Key_Menu      =0x01000055
    Key_Hyper_L   =0x01000056
    Key_Hyper_R   =0x01000057
    Key_Help      =0x01000058
    Key_Direction_L =0x01000059
    Key_Direction_R =0x01000060
    Key_Space     =0x20
    Key_Any       = Key_Space
    Key_Exclam    = 0x21
    Key_QuoteDbl  =    0x22
    Key_NumberSign    =  0x23
    Key_Dollar        = 0x24
    Key_Percent       = 0x25
    Key_Ampersand     =  0x26
    Key_Apostrophe    =  0x27
    Key_ParenLeft     =  0x28
    Key_ParenRight    =  0x29
    Key_Asterisk      =  0x2a
    Key_Plus          = 0x2b
    Key_Comma         = 0x2c
    Key_Minus         = 0x2d
    Key_Period        = 0x2e
    Key_Slash         = 0x2f
    Key_0             = 0x30
    Key_1             = 0x31
    Key_2             = 0x32
    Key_3             = 0x33
    Key_4             = 0x34
    Key_5             = 0x35
    Key_6             = 0x36
    Key_7             = 0x37
    Key_8             = 0x38
    Key_9             = 0x39
    Key_Colon         = 0x3a
    Key_Semicolon     = 0x3b
    Key_Less          = 0x3c
    Key_Equal         = 0x3d
    Key_Greater       = 0x3e
    Key_Question      = 0x3f
    Key_At            = 0x40
    Key_A             = 0x41
    Key_B             = 0x42
    Key_C             = 0x43
    Key_D             = 0x44
    Key_E             = 0x45
    Key_F             = 0x46
    Key_G             = 0x47
    Key_H             = 0x48
    Key_I             = 0x49
    Key_J             = 0x4a
    Key_K             = 0x4b
    Key_L             = 0x4c
    Key_M             = 0x4d
    Key_N             = 0x4e
    Key_O             = 0x4f
    Key_P             = 0x50
    Key_Q             = 0x51
    Key_R             = 0x52
    Key_S             = 0x53
    Key_T             = 0x54
    Key_U             = 0x55
    Key_V             = 0x56
    Key_W             = 0x57
    Key_X             = 0x58
    Key_Y             = 0x59
    Key_Z             = 0x5a
    Key_BracketLeft =0x5b
    Key_Backslash   =0x5c
    Key_BracketRight =   0x5d
    Key_AsciiCircum =0x5e
    Key_Underscore  =0x5f
    Key_QuoteLeft   =0x60
    Key_BraceLeft   =0x7b
    Key_Bar         =0x7c
    Key_BraceRight  =0x7d
    Key_AsciiTilde  =0x7e
    Key_unknown     = 0x01FFFFFF

    NonPrintableMask = 0x01000000
    Mask             = 0x01FFFFFF

class KeyModifier:
    NoModifier      = 0x00000000
    ShiftModifier   = 0x02000000
    ControlModifier = 0x04000000
    AltModifier     = 0x08000000
    MetaModifier    = 0x10000000
    KeypadModifier  = 0x20000000

    Mask            = 0x3F000000

def nativeToVaiKeyCode(native_key_code):
    """Transforms the keycode from native (ncurses) to vaitk representation"""
    key_mapper = {
      1                    : Key.Key_A | KeyModifier.ControlModifier,
      2                    : Key.Key_B | KeyModifier.ControlModifier,
      3                    : Key.Key_C | KeyModifier.ControlModifier,
      4                    : Key.Key_D | KeyModifier.ControlModifier,
      5                    : Key.Key_E | KeyModifier.ControlModifier,
      6                    : Key.Key_F | KeyModifier.ControlModifier,
      7                    : Key.Key_G | KeyModifier.ControlModifier,
      8                    : Key.Key_Backspace,
      9                    : Key.Key_Tab,
      10                   : Key.Key_Return,
      11                   : Key.Key_K | KeyModifier.ControlModifier,
      12                   : Key.Key_L | KeyModifier.ControlModifier,
      13                   : Key.Key_M | KeyModifier.ControlModifier,
      14                   : Key.Key_N | KeyModifier.ControlModifier,
      15                   : Key.Key_O | KeyModifier.ControlModifier,
      16                   : Key.Key_P | KeyModifier.ControlModifier,
      17                   : Key.Key_Q | KeyModifier.ControlModifier,
      18                   : Key.Key_R | KeyModifier.ControlModifier,
      19                   : Key.Key_S | KeyModifier.ControlModifier,
      20                   : Key.Key_T | KeyModifier.ControlModifier,
      21                   : Key.Key_U | KeyModifier.ControlModifier,
      22                   : Key.Key_V | KeyModifier.ControlModifier,
      23                   : Key.Key_W | KeyModifier.ControlModifier,
      24                   : Key.Key_X | KeyModifier.ControlModifier,
      25                   : Key.Key_Y | KeyModifier.ControlModifier,
      26                   : Key.Key_Z | KeyModifier.ControlModifier,
      27                   : Key.Key_Escape,
      32                   : Key.Key_Space,
      33                   : Key.Key_Exclam,
      34                   : Key.Key_QuoteDbl,
      35                   : Key.Key_NumberSign,
      36                   : Key.Key_Dollar,
      37                   : Key.Key_Percent,
      38                   : Key.Key_Ampersand,
      39                   : Key.Key_Apostrophe,
      40                   : Key.Key_ParenLeft,
      41                   : Key.Key_ParenRight,
      42                   : Key.Key_Asterisk,
      43                   : Key.Key_Plus,
      44                   : Key.Key_Comma,
      45                   : Key.Key_Minus,
      46                   : Key.Key_Period,
      47                   : Key.Key_Slash,
      48                   : Key.Key_0,
      49                   : Key.Key_1,
      50                   : Key.Key_2,
      51                   : Key.Key_3,
      52                   : Key.Key_4,
      53                   : Key.Key_5,
      54                   : Key.Key_6,
      55                   : Key.Key_7,
      56                   : Key.Key_8,
      57                   : Key.Key_9,
      58                   : Key.Key_Colon,
      59                   : Key.Key_Semicolon,
      60                   : Key.Key_Less,
      61                   : Key.Key_Equal,
      62                   : Key.Key_Greater,
      63                   : Key.Key_Question,
      64                   : Key.Key_At,
      65                   : Key.Key_A | KeyModifier.ShiftModifier,
      66                   : Key.Key_B | KeyModifier.ShiftModifier,
      67                   : Key.Key_C | KeyModifier.ShiftModifier,
      68                   : Key.Key_D | KeyModifier.ShiftModifier,
      69                   : Key.Key_E | KeyModifier.ShiftModifier,
      70                   : Key.Key_F | KeyModifier.ShiftModifier,
      71                   : Key.Key_G | KeyModifier.ShiftModifier,
      72                   : Key.Key_H | KeyModifier.ShiftModifier,
      73                   : Key.Key_I | KeyModifier.ShiftModifier,
      74                   : Key.Key_J | KeyModifier.ShiftModifier,
      75                   : Key.Key_K | KeyModifier.ShiftModifier,
      76                   : Key.Key_L | KeyModifier.ShiftModifier,
      77                   : Key.Key_M | KeyModifier.ShiftModifier,
      78                   : Key.Key_N | KeyModifier.ShiftModifier,
      79                   : Key.Key_O | KeyModifier.ShiftModifier,
      80                   : Key.Key_P | KeyModifier.ShiftModifier,
      81                   : Key.Key_Q | KeyModifier.ShiftModifier,
      82                   : Key.Key_R | KeyModifier.ShiftModifier,
      83                   : Key.Key_S | KeyModifier.ShiftModifier,
      84                   : Key.Key_T | KeyModifier.ShiftModifier,
      85                   : Key.Key_U | KeyModifier.ShiftModifier,
      86                   : Key.Key_V | KeyModifier.ShiftModifier,
      87                   : Key.Key_W | KeyModifier.ShiftModifier,
      88                   : Key.Key_X | KeyModifier.ShiftModifier,
      89                   : Key.Key_Y | KeyModifier.ShiftModifier,
      90                   : Key.Key_Z | KeyModifier.ShiftModifier,
      91                   : Key.Key_BracketLeft,
      92                   : Key.Key_Backslash,
      93                   : Key.Key_BracketRight,
      94                   : Key.Key_AsciiCircum,
      95                   : Key.Key_Underscore,
      96                   : Key.Key_QuoteLeft,
      97                   : Key.Key_A,
      98                   : Key.Key_B,
      99                   : Key.Key_C,
      100                  : Key.Key_D,
      101                  : Key.Key_E,
      102                  : Key.Key_F,
      103                  : Key.Key_G,
      104                  : Key.Key_H,
      105                  : Key.Key_I,
      106                  : Key.Key_J,
      107                  : Key.Key_K,
      108                  : Key.Key_L,
      109                  : Key.Key_M,
      110                  : Key.Key_N,
      111                  : Key.Key_O,
      112                  : Key.Key_P,
      113                  : Key.Key_Q,
      114                  : Key.Key_R,
      115                  : Key.Key_S,
      116                  : Key.Key_T,
      117                  : Key.Key_U,
      118                  : Key.Key_V,
      119                  : Key.Key_W,
      120                  : Key.Key_X,
      121                  : Key.Key_Y,
      122                  : Key.Key_Z,
      123                  : Key.Key_BraceLeft,
      124                  : Key.Key_Bar,
      125                  : Key.Key_BraceRight,
      126                  : Key.Key_AsciiTilde,
      127                  : Key.Key_Backspace,
      curses.KEY_DOWN      : Key.Key_Down,
      curses.KEY_UP        : Key.Key_Up,
      curses.KEY_LEFT      : Key.Key_Left,
      curses.KEY_RIGHT     : Key.Key_Right,
      curses.KEY_BACKSPACE : Key.Key_Backspace,
      curses.KEY_NPAGE     : Key.Key_PageDown,
      curses.KEY_PPAGE     : Key.Key_PageUp,
      curses.KEY_DC        : Key.Key_Delete,
    }
    return key_mapper.get(native_key_code)

def isKeyCodePrintable(key_code):
    return ((key_code & Key.NonPrintableMask) == 0)

def vaiKeyCodeToText(key_code):
    if not isKeyCodePrintable(key_code):
        return ''

    key_map = {
        Key.Key_Space        : ' ',
        Key.Key_Exclam       : '!',
        Key.Key_QuoteDbl     : '"',
        Key.Key_NumberSign   : '#',
        Key.Key_Dollar       : '$',
        Key.Key_Percent      : '%',
        Key.Key_Ampersand    : '&',
        Key.Key_Apostrophe   : "'",
        Key.Key_ParenLeft    : '(',
        Key.Key_ParenRight   : ')',
        Key.Key_Asterisk     : '*',
        Key.Key_Plus         : '+',
        Key.Key_Comma        : ',',
        Key.Key_Minus        : '-',
        Key.Key_Period       : '.',
        Key.Key_Slash        : '/',
        Key.Key_0            : '0',
        Key.Key_1            : '1',
        Key.Key_2            : '2',
        Key.Key_3            : '3',
        Key.Key_4            : '4',
        Key.Key_5            : '5',
        Key.Key_6            : '6',
        Key.Key_7            : '7',
        Key.Key_8            : '8',
        Key.Key_9            : '9',
        Key.Key_Colon        : ':',
        Key.Key_Semicolon    : ';',
        Key.Key_Less         : '<',
        Key.Key_Equal        : '=',
        Key.Key_Greater      : '>',
        Key.Key_Question     : '?',
        Key.Key_At           : '@',
        Key.Key_A            : 'a',
        Key.Key_B            : 'b',
        Key.Key_C            : 'c',
        Key.Key_D            : 'd',
        Key.Key_E            : 'e',
        Key.Key_F            : 'f',
        Key.Key_G            : 'g',
        Key.Key_H            : 'h',
        Key.Key_I            : 'i',
        Key.Key_J            : 'j',
        Key.Key_K            : 'k',
        Key.Key_L            : 'l',
        Key.Key_M            : 'm',
        Key.Key_N            : 'n',
        Key.Key_O            : 'o',
        Key.Key_P            : 'p',
        Key.Key_Q            : 'q',
        Key.Key_R            : 'r',
        Key.Key_S            : 's',
        Key.Key_T            : 't',
        Key.Key_U            : 'u',
        Key.Key_V            : 'v',
        Key.Key_W            : 'w',
        Key.Key_X            : 'x',
        Key.Key_Y            : 'y',
        Key.Key_Z            : 'z',
        Key.Key_BracketLeft  : '[',
        Key.Key_Backslash    : '\\',
        Key.Key_BracketRight : ']',
        Key.Key_AsciiCircum  : '^',
        Key.Key_Underscore   : '_',
        Key.Key_QuoteLeft    : '`',
        Key.Key_BraceLeft    : '{',
        Key.Key_Bar          : '|',
        Key.Key_BraceRight   : '}',
        Key.Key_AsciiTilde   : '~',

        Key.Key_A | KeyModifier.ShiftModifier : 'A',
        Key.Key_B | KeyModifier.ShiftModifier : 'B',
        Key.Key_C | KeyModifier.ShiftModifier : 'C',
        Key.Key_D | KeyModifier.ShiftModifier : 'D',
        Key.Key_E | KeyModifier.ShiftModifier : 'E',
        Key.Key_F | KeyModifier.ShiftModifier : 'F',
        Key.Key_G | KeyModifier.ShiftModifier : 'G',
        Key.Key_H | KeyModifier.ShiftModifier : 'H',
        Key.Key_I | KeyModifier.ShiftModifier : 'I',
        Key.Key_J | KeyModifier.ShiftModifier : 'J',
        Key.Key_K | KeyModifier.ShiftModifier : 'K',
        Key.Key_L | KeyModifier.ShiftModifier : 'L',
        Key.Key_M | KeyModifier.ShiftModifier : 'M',
        Key.Key_N | KeyModifier.ShiftModifier : 'N',
        Key.Key_O | KeyModifier.ShiftModifier : 'O',
        Key.Key_P | KeyModifier.ShiftModifier : 'P',
        Key.Key_Q | KeyModifier.ShiftModifier : 'Q',
        Key.Key_R | KeyModifier.ShiftModifier : 'R',
        Key.Key_S | KeyModifier.ShiftModifier : 'S',
        Key.Key_T | KeyModifier.ShiftModifier : 'T',
        Key.Key_U | KeyModifier.ShiftModifier : 'U',
        Key.Key_V | KeyModifier.ShiftModifier : 'V',
        Key.Key_W | KeyModifier.ShiftModifier : 'W',
        Key.Key_X | KeyModifier.ShiftModifier : 'X',
        Key.Key_Y | KeyModifier.ShiftModifier : 'Y',
        Key.Key_Z | KeyModifier.ShiftModifier : 'Z',
    }

    return key_map.get(key_code, '')





from vaitk import gui
from pygments import token
import collections

class Icons:
    _ICONS = None


    @classmethod
    def getCollection(cls, collection_name):
        if cls._ICONS is None:
            cls._initIcons()
        return cls._ICONS[collection_name]

    @classmethod
    def _initIcons(cls):
        cls._ICONS = {}
        cls._ICONS["ascii"] = {
                                 "EditArea.tab" : ".",
                                 "SideRuler.warning" : "W",
                                 "SideRuler.error"   : "E",
                                 "SideRuler.error"   : "I",
                                 "SideRuler.added"   : "+",
                                 "SideRuler.deletion_before" : "_",
                                 "SideRuler.deletion_after"  : "^",
                                 "SideRuler.modified"        : ".",
                                 "SideRuler.unexistent_line" : "~",
                                 "SideRuler.border"          : "|",
                                 "SideRuler.bookmarks"       : "abcdefghijklmnopqrstuvwxyz"
                                }

        cls._ICONS["unicode1"] = {
                                 "EditArea.tab"              : "\u254e", # dotted lines
                                 "SideRuler.warning"         : "\u203c", # Double exclamation
                                 "SideRuler.error"           : "\u203c", # Double exclamation
                                 "SideRuler.info"            : "I",
                                 "SideRuler.added"           : "+",
                                 "SideRuler.deletion_before" : "\u2304", # Arrow
                                 "SideRuler.deletion_after"  : "\u2303", # Arrow
                                 "SideRuler.modified"        : "\u25fe", # Square
                                 "SideRuler.unexistent_line" : "~",
                                 "SideRuler.border"          : "\u2551", # double line
                                 "SideRuler.bookmarks"       :
                                                               "\u24D0"+
                                                               "\u24D1"+
                                                               "\u24D2"+
                                                               "\u24D3"+
                                                               "\u24D4"+
                                                               "\u24D5"+
                                                               "\u24D6"+
                                                               "\u24D7"+
                                                               "\u24D8"+
                                                               "\u24D9"+
                                                               "\u24DA"+
                                                               "\u24DB"+
                                                               "\u24DC"+
                                                               "\u24DD"+
                                                               "\u24DE"+
                                                               "\u24DF"+
                                                               "\u24E0"+
                                                               "\u24E1"+
                                                               "\u24E2"+
                                                               "\u24E3"+
                                                               "\u24E4"+
                                                               "\u24E5"+
                                                               "\u24E6"+
                                                               "\u24E7"+
                                                               "\u24E8"+
                                                               "\u24E9"+
                                                               "\u24B6"+
                                                               "\u24B7"+
                                                               "\u24B8"+
                                                               "\u24B9"+
                                                               "\u24BA"+
                                                               "\u24BB"+
                                                               "\u24BC"+
                                                               "\u24BD"+
                                                               "\u24BE"+
                                                               "\u24BF"+
                                                               "\u24C0"+
                                                               "\u24C1"+
                                                               "\u24C2"+
                                                               "\u24C3"+
                                                               "\u24C4"+
                                                               "\u24C5"+
                                                               "\u24C6"+
                                                               "\u24C7"+
                                                               "\u24C8"+
                                                               "\u24C9"+
                                                               "\u24CA"+
                                                               "\u24CB"+
                                                               "\u24CC"+
                                                               "\u24CD"+
                                                               "\u24CE"+
                                                               "\u24CF"
                                }





from ..EditorApp import EditorApp
from vai.lexer import token as lexertoken
from vaitk import gui

def application():
    """
    Returns the Vai application.
    """

    return EditorApp.vApp

def statusBar():
    """
    Returns the application's status bar.
    """
    return application().editor.status_bar




#!/usr/bin/env python3
__version__ = "1.6"

from vaitk import core, gui
from . import EditorApp
from . import models
from . import BugReport
import os
import sys
import io
import argparse
import locale
import logging

def _workaroundNCurses():
    """
    This routine works around the ncurses library.
    It must be called before everything else, so that ncurses is not
    loaded.
    The problem is that python installs and links by default to
    ncurses, instead of ncursesw (wide characters). The result is that unicode
    stuff gets corrupted. We work around the problem as explained in the
    comments. The two libraries appear to be compatible, so if I just "mimic"
    the library name, it behaves correctly.

    Unfortunately, we _must_ rely on a shell script to export LD_LIBRARY_PATH
    for us. I found out that dlopen does not honor LD_LIBRARY_PATH if changed
    after startup.

    The end result is that I always copy the ncurses wide character when not
    in a local lib source.
    """

    import platform
    import os
    import sys

    if platform.system() != "Linux":
        return

    local_lib_path = os.path.expanduser("~/.local/lib/vai/")

    # We already deployed the workaround
    if os.path.exists(os.path.join(local_lib_path, "libncurses.so")):
        return

    if platform.processor() == "x86_64":
        ncurses_path = "/lib/x86_64-linux-gnu/libncursesw.so.5.9"
    else:
        ncurses_path = "/lib/i386-linux-gnu/libncursesw.so.5.9"

    # If the library does not exist on the system, we give up and start rendering in ascii
    if not os.path.exists(ncurses_path):
        from . import models
        models.Configuration.flags["has_wide_ncurses"] = False
        return

    # If present, copy the libraries to the local lib path, with a different name, to trick
    # python in using them
    import shutil
    os.makedirs(local_lib_path, exist_ok=True)

    # Just three copies, I'm a lazy bum.
    shutil.copyfile(ncurses_path, os.path.join(local_lib_path, "libncurses.so.5.9"))
    shutil.copyfile(ncurses_path, os.path.join(local_lib_path, "libncurses.so.5"))
    shutil.copyfile(ncurses_path, os.path.join(local_lib_path, "libncurses.so"))

    # Signals the shell script that the thing has been deployed and need to restart vai
    sys.exit(42)

def main():
    _workaroundNCurses()
    # Sets the locale to the default (user specified) locale, as suggested by ncurses docs
    locale.setlocale(locale.LC_ALL, "")

    parser = argparse.ArgumentParser(description="A Vim-like editor, with big dreams.")
    parser.add_argument("filename", nargs="?", help="The filename to open")
    parser.add_argument("--dump-default-config",
                        help="Dump the default configuration to the config file. Useful to reset a broken configuration.",
                        action="store_true")
    parser.add_argument("--profile", help="Enable profiling at the end of the run.", action="store_true")
    parser.add_argument("--pdb", help="Enable debugging with pdb in case of crash.", action="store_true")
    parser.add_argument("--noreport", help="Skip request for bug reporting.", action="store_true")
    parser.add_argument("--version", '-v', help="Print the version number.", action="version",
                                           version='vai {0}'.format(__version__))
    args = parser.parse_args()

    app = None
    try:

        if args.dump_default_config:
            filename = models.Configuration.filename()
            if os.path.exists(filename):
                print("Refusing to overwrite existing config file %s. Delete the file manually and try again." % filename)
                sys.exit(1)

            models.Configuration.save()
            print("Dumped default configuration in %s" % filename)
            sys.exit(0)
        app = EditorApp.EditorApp(sys.argv)
        if args.filename:
            app.openFile(args.filename)

        if args.profile:
            import cProfile
            pr = cProfile.Profile()
            pr.enable()

        app.exec_()

        if args.profile:
            pr.disable()
            s = io.StringIO()
            sortby = 'tottime'
            import pstats
            ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
            ps.print_stats()
            print(s.getvalue())

        app.resetScreen()

    except Exception as e:
        import traceback
        import contextlib

        saved_files = []
        if app is not None:
            saved_files = app.dumpBuffers()

            app.resetScreen()

        with contextlib.closing(open("vai_crashreport.out", "w")) as f:
            f.write(traceback.format_exc())

        if args.pdb:
            import pdb
            pdb.post_mortem()

        if not args.noreport:
            BugReport.report(saved_files)

    return 0




from yapsy.IPlugin import IPlugin
from ..models import SyntaxColors

class SyntaxColorsPlugin(IPlugin):
    def getSchema(self, num_colors):
        """
        To be reimplement in the plugin to return a new schema.
        """
        return None

    def name(self):
        """
        To be reimplement in the plugin
        """
        return None

    def supportsNumColors(self, num_colors):
        """
        To be reimplement in the plugin
        """
        return False

    def _defaultSchema(num_colors=None):
        """
        Returns the default color schema for a given number of colors,
        or for the current number of colors if not specified
        """
        num_colors = num_colors or self.numColors()
        return SyntaxColors.defaultColorSchema(num_colors)



import vaitk
from vaitk import core, gui

from .models import EditorMode

class EditAreaEventFilter(core.VObject):
    """
    Event filter to detect the use of commandbar initiation
    keys, such as :, / and ?
    """
    def __init__(self, command_bar, global_state, buffer_list):
        super().__init__()
        self._command_bar = command_bar
        self._global_state = global_state
        self._buffer_list = buffer_list

    def eventFilter(self, event):
        if not isinstance(event, gui.VKeyEvent):
            return False

        if self._global_state.editor_mode != EditorMode.COMMAND:
            return False

        if event.key() == vaitk.Key.Key_Colon:
            self._global_state.editor_mode = EditorMode.COMMAND_INPUT
            self._command_bar.setFocus()
            return True

        if event.key() == vaitk.Key.Key_Slash:
            self._global_state.editor_mode = EditorMode.SEARCH_FORWARD
            self._command_bar.setFocus()
            return True

        if event.key() == vaitk.Key.Key_Question:
            self._global_state.editor_mode = EditorMode.SEARCH_BACKWARD
            self._command_bar.setFocus()
            return True

        if event.key() == vaitk.Key.Key_N and event.modifiers() & vaitk.KeyModifier.ControlModifier:
            self._buffer_list.selectNext()
            return True

        if event.key() == vaitk.Key.Key_P and event.modifiers() & vaitk.KeyModifier.ControlModifier:
            self._buffer_list.selectPrev()
            return True

        return False





from ... import core
from ... import Key, KeyModifier, nativeToVaiKeyCode, vaiKeyCodeToText
import logging
import sys
import os
import copy
import select
from ..VApplication import VApplication
from ..VColor import VColor
from ..VPalette import VPalette
from ..VScreen import VScreen
from ..VPainter import VPainter
from ..VWidget import VWidget

from .VFrame import VFrame
from .VDialog import VDialog
from .VLabel import VLabel
from .VToolTip import VToolTip
from .VLineEdit import VLineEdit
from .VPushButton import VPushButton
from .VTabWidget import VTabWidget



from .BufferCommand import BufferCommand
from .CommandResult import CommandResult
from .NewLineCommand import NewLineCommand
from .NewLineAfterCommand import NewLineAfterCommand

class BreakLineCommand(BufferCommand):
    def execute(self):
        cursor = self._cursor
        document = self._document
        if self.savedCursorPos() is None:
            self.saveCursorPos()
        
        pos = self.savedCursorPos()

        if pos[1] == document.lineLength(pos[0]):
            command = NewLineAfterCommand(self._buffer)
            result = command.execute()
            if result.success:
                self._sub_command = command
            return result

        if pos[1] == 1:
            command = NewLineCommand(self._buffer)
            result = command.execute()
            if result.success:
                self._sub_command = command
            cursor.toPos((pos[0]+1, 1))
            return result

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        current_text = document.lineText(pos[0])
        current_indent = len(current_text) - len(current_text.lstrip(' '))

        document.breakLine(pos)
        text_broken_line = document.lineText(pos[0]+1)
        document.deleteChars((pos[0]+1, 1), len(text_broken_line)-len(text_broken_line.lstrip(' ')))
        document.insertChars( (pos[0]+1, 1), ' '*current_indent )
        cursor.toPos((pos[0]+1, current_indent+1))

        line_meta = document.lineMetaInfo("Change")

        if line_meta.data(pos[0]) == None:
            line_meta.setData([ "modified", "added" ], pos[0])
        else:
            line_meta.setData([ "added" ], pos[0]+1)

        return CommandResult(success=True, info=None)

    def undo(self):
        self.restoreCursorPos()

        if self._sub_command is not None:
            self._sub_command.undo()
            self._sub_command = None
            return

        self.restoreLineMemento()
        self._document.deleteLine(self._cursor.pos[0]+1)




from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class NewLineCommand(BufferCommand):
    def execute(self):
        document = self._document
        cursor = self._cursor

        if self.savedCursorPos() is None:
            self.saveCursorPos()
        
        pos = self.savedCursorPos()

        current_text = document.lineText(pos[0])
        current_indent = len(current_text) - len(current_text.lstrip(' '))

        document.newLine(pos[0])
        document.insertChars( (pos[0], 1), ' '*current_indent )
        line_meta = document.lineMetaInfo("Change")
        line_meta.setData("added", pos[0])
        cursor.toPos( (pos[0], document.lineLength(pos[0])))

        return CommandResult(True, None)

    def undo(self):
        self._document.deleteLine(self.savedCursorPos()[0])
        self.restoreCursorPos()




from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class DeleteToEndOfLineCommand(BufferCommand):
    def execute(self):
        cursor = self._buffer.cursor
        document = self._buffer.document

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()
        cursor.toPos(pos)

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        line_meta = document.lineMetaInfo("Change")
        changed = line_meta.data(pos[0])
        if changed:
            line_meta.setData("modified", pos[0])

        deleted = document.deleteChars(pos, document.lineLength(pos[0])-pos[1])
        cursor.toCharPrev()
        return CommandResult(success=True, info=deleted)




from ..VWidget import VWidget
class VTabWidget(VWidget):
    def __init__(self, parent=None):
        super(VTabWidget, self).__init__(parent)
        self._tabs = []
        self._selected_tab_idx = -1

    def addTab(self, widget, label):
        self._tabs.append((widget, label))
        self._selected_tab_idx = 2

    def render(self, screen):
        screen = VApplication.vApp.screen()
        w, h = screen.size()
        if len(self._tabs):
            tab_size = w/len(self._tabs)
            header = ""
            for index, (_, label) in enumerate(self._tabs):
                header = label+" "*(tab_size-len(label))
                screen.write(tab_size * index, 0, header, curses.color_pair(1 if index == self._selected_tab_idx else 0))
            widget = self._tabs[self._selected_tab_idx][0]
            widget.render(screen)




from pygments import lexers, util
from ..SymbolLookupDb import SymbolLookupDb
from ..models.TextDocument import CharMeta
from . import token
import os

# Faster lookup than the one provided in lexers.get_lexer_for_filename.
# Any other attempt has been proved to take around 0.2 secs, which is
# unacceptable. We need the fastest lookup to start as soon as possible
EXTENSION_TO_LEXER = {
    ".py" : lexers.PythonLexer,
    ".rb" : lexers.RubyLexer,
    ".pl" : lexers.PerlLexer,
    ".tcl": lexers.TclLexer,
    ".lua": lexers.LuaLexer,
    ".as" : lexers.ActionScriptLexer,
    ".c"  : lexers.CLexer,
    ".h"  : lexers.CLexer,
    ".cpp" : lexers.CppLexer,
    ".hpp" : lexers.CppLexer,
    ".f"   : lexers.FortranLexer,
    ".f77" : lexers.FortranLexer,
    ".f90" : lexers.FortranLexer,
    ".sh"  : lexers.BashLexer,
    ".bat" : lexers.BatchLexer,
}

def _getLexerInstance(filename):
    """Get the lexer instance from the filename"""
    if filename is None:
        return lexers.TextLexer(stripnl=False, stripall=False)

    # Fast lookup first
    ext = os.path.splitext(filename)[1]
    cls = EXTENSION_TO_LEXER.get(ext)
    if cls is not None:
        return cls(stripnl=False, stripall=False)

    # Slow lookup for less common cases
    try:
        lexer = lexers.get_lexer_for_filename(filename, stripnl=False, stripall=False)
    except util.ClassNotFound:
        lexer = lexers.TextLexer(stripnl=False, stripall=False)

    return lexer


class Lexer:
    """
    Observes the TextDocument for changes, and performs lexing
    of its contents synchronously. The text is parsed with the
    lexer as specified by the document meta information FileType.
    """
    def __init__(self):
        self._document = None
        self._lexer = None

    def setModel(self, document):
        """Sets the textdocument as a model for the lexer"""
        self._document = document
        filename = self._document.filename()
        self._lexer = _getLexerInstance(filename)
        self._document.contentChanged.connect(self._lexContents)
        file_type_meta = self._document.documentMetaInfo("FileType")
        if file_type_meta.data() is None:
            file_type_meta.setData(self._lexer.name)
        self._lexContents()

    def _lexContents(self):
        """
        Perform lexing of the document every time it changes.
        Fills the meta information on the document.
        """

        if self._lexer is None:
            return

        tokens = self._lexer.get_tokens(self._document.documentText())
        current_line = 1
        current_col = 1
        SymbolLookupDb.clear()
        # Skip the space token

        tokens = self._processTokens(tokens)
        for tok in tokens:
            ttype, token_string = tok
            if ttype in [token.Name, token.Name.Class, token.Name.Function]:
                SymbolLookupDb.add(token_string)

            ttype = self._processToken(ttype, token_string)

            token_lines = token_string.splitlines(True)
            for token_line in token_lines:
                meta = [ttype]*len(token_line)

                self._document.updateCharMeta((current_line, current_col), {CharMeta.LexerToken: meta})
                current_col += len(token_line)
                if token_line.endswith("\n"):
                    current_line += 1
                    current_col = 1

    def _processToken(self, ttype, token_string):
        if token_string.startswith("__") and token_string.endswith("__") and ttype is token.Name.Function:
            return token.Name.Function.PythonMagic
        elif token_string.startswith("_"):
            if ttype is token.Name.Class:
                return token.Name.Class.PythonPrivate
            elif ttype is token.Name.Function:
                return token.Name.Function.PythonPrivate
        elif token_string == "self":
            return token.Name.Builtin.Pseudo.PythonSelf

        return ttype

    def _processTokens(self, tokens):
        return tokens



from vaitk import gui

from . import widgets
from . import controllers

from .EditArea import EditArea
from .EditAreaEventFilter import EditAreaEventFilter
from .InfoHoverBox import InfoHoverBox


class Editor(gui.VWidget):
    """
    Widget responsible for handling the overall aspect of the editor,
    aggregating the different components.
    """

    def __init__(self, editor_app, global_state, buffer_list, parent=None):
        super().__init__(parent=parent)

        self._editor_app = editor_app
        self._global_state = global_state
        self._buffer_list = buffer_list
        self._controller = controllers.EditorController(self, self._global_state, self._buffer_list)

        self._createStatusBar()
        self._createCommandBar()
        self._createSideRuler()
        self._createEditArea()
        self._createInfoHoverBox()

        self._status_bar_controller = controllers.StatusBarController(self._status_bar)
        self._side_ruler_controller = controllers.SideRulerController(self._side_ruler)
        self._command_bar_controller = controllers.CommandBarController(self._command_bar, self._edit_area, self._controller, self._global_state)
        self._edit_area_event_filter = EditAreaEventFilter(self._command_bar, self._global_state, self._buffer_list)
        self._edit_area.installEventFilter(self._edit_area_event_filter)

        self._controller.registerCurrentBuffer()

    def show(self):
        super().show()
        self._edit_area.setFocus()

    # properties

    @property
    def editor_app(self):
        return self._editor_app

    @property
    def status_bar(self):
        return self._status_bar

    @property
    def edit_area(self):
        return self._edit_area

    @property
    def status_bar_controller(self):
        return self._status_bar_controller

    @property
    def side_ruler_controller(self):
        return self._side_ruler_controller

    @property
    def controller(self):
        return self._controller

    @property
    def info_hover_box(self):
        return self._info_hover_box

    # Private

    def _createStatusBar(self):
        self._status_bar = widgets.StatusBar(self)
        self._status_bar.move( (0, self.height()-2) )
        self._status_bar.resize( (self.width(), 1) )

    def _createCommandBar(self):
        self._command_bar = widgets.CommandBar(self)
        self._command_bar.move( (0, self.height()-1) )
        self._command_bar.resize( (self.width(), 1) )

    def _createSideRuler(self):
        self._side_ruler = widgets.SideRuler(self)
        self._side_ruler.move( (0, 0) )
        self._side_ruler.resize( (7, self.height()-2) )

    def _createEditArea(self):
        self._edit_area = EditArea(self._global_state, self._controller, parent = self)
        self._edit_area.move( (7, 0) )
        self._edit_area.resize((self.width()-4, self.height()-2) )

    def _createInfoHoverBox(self):
        self._info_hover_box = InfoHoverBox()




from .BufferList import BufferList
from .Buffer import Buffer
from .EditAreaModel import EditAreaModel
from .GlobalState import GlobalState
from .TextDocumentCursor import TextDocumentCursor
from .TextDocument import TextDocument
from .EditorMode import EditorMode
from .Configuration import Configuration
from .EditorState import EditorState
from .SyntaxColors import SyntaxColors
from .Icons import Icons



import re

class SearchDirection:
    FORWARD = 1
    BACKWARD = -1

def findAll(document, search_text, line_interval=None, case_sensitive=True, word=False):
    """
    Find all occurrences of a given search text (evt. regexp text).
    """
    match_pos = []
    if line_interval is None:
        line_interval = (1, document.numLines()+1)
    else:
        line_interval = (max(1, line_interval[0]), min(document.numLines()+1, line_interval[1]))

    flags = 0
    if not case_sensitive:
       flags = re.IGNORECASE

    if word:
        search_text = r'\b'+re.escape(search_text)+r'\b'
    else:
        search_text = re.escape(search_text)

    allMatches = re.compile(search_text, flags).finditer

    for line_num in range(*line_interval):
        line_text = document.lineText(line_num)
        match_pos.extend( ((line_num, m.start()+1, m.end()+1) for m in allMatches(line_text)) )

    return match_pos

def find(buffer, text, direction):
    """
    Find the next occurrence of text in the buffer, with a given direction.
    If found, it moves the cursor to the appropriate position and returns True
    If not found, it returns false.
    """

    document = buffer.document
    cursor = buffer.cursor

    pos = cursor.pos

    current_line, current_col = pos

    # We do either forward or reverse search all with the same code.
    # We just switch the lookup function depending on the direction.
    # and apply it to each line string.
    if direction == SearchDirection.FORWARD:
        first_half, second_half = ((pos[0], document.numLines()+1), (1, pos[0]+1))
        find_routine = "find"
    else:
        first_half, second_half = ((pos[0], 0, -1), (document.numLines(), pos[0]-1, -1))
        find_routine = "rfind"


    # Perform the lookup in the first half.
    for line_num in range(*first_half):
        start, stop = (None, None)
        if line_num == current_line:
            # current_col - 1 + 1 = next char
            start, stop = (current_col, None) if direction == SearchDirection.FORWARD else (None, current_col)

        index = getattr(document.lineText(line_num), find_routine)(text, start, stop)
        if index != -1:
            cursor.toPos((line_num, index+1))
            return True

    # Reached the end, start from the other side
    for line_num in range(*second_half):
        start, stop = (None, None)
        if line_num == current_line:
            start, stop = (None, current_col) if direction == SearchDirection.FORWARD else (current_col, None)
        else:
            start, stop = (0, None)

        index = getattr(document.lineText(line_num), find_routine)(text, start, stop)

        if index != -1:
            cursor.toPos((line_num, index+1))
            return True

    return False




from .. import Key, KeyModifier, nativeToVaiKeyCode, vaiKeyCodeToText
from ..core import VEvent

class VKeyEvent(VEvent):
    def __init__(self, key_code):
        super().__init__(VEvent.EventType.KeyPress)
        self._key_code = key_code
        self._accepted = False

    def keyCode(self):
        return self._key_code

    def key(self):
        return self._key_code & Key.Mask

    def modifiers(self):
        return self._key_code & KeyModifier.Mask

    def text(self):
        return vaiKeyCodeToText(self._key_code)

    @staticmethod
    def fromNativeKeyCode(native_key_code):
        key_code = nativeToVaiKeyCode(native_key_code)
        if key_code is None:
            return None
        return VKeyEvent(key_code)

class VFocusEvent(VEvent):
    def __init__(self, focus_type):
        super().__init__(focus_type)

class VPaintEvent(VEvent):
    def __init__(self):
        super().__init__(VEvent.EventType.Paint)

class VHideEvent(VEvent):
    def __init__(self):
        super().__init__(VEvent.EventType.Hide)

class VShowEvent(VEvent):
    def __init__(self):
        super().__init__(VEvent.EventType.Show)

class VMoveEvent(VEvent):
    def __init__(self):
        super().__init__(VEvent.EventType.Move)

class VResizeEvent(VEvent):
    def __init__(self):
        super().__init__(VEvent.EventType.Resize)




from vaitk import core

class TextDocumentCursor(core.VObject):
    def __init__(self, text_document):
        self._text_document = text_document
        self._text_document.registerCursor(self)
        self._pos = (1,1)
        self._optimistic_column = 1

        self.positionChanged = core.VSignal(self)

    def textDocument(self):
        return self._text_document

    @property
    def pos(self):
        return self._pos

    @property
    def line(self):
        return self._pos[0]

    @property
    def column(self):
        return self._pos[1]

    # Note: not a setter, because we want to check success in a "soft" way.
    # at least for now.
    def toPos(self, pos):
        """
        Move to a specific position if possible. Return True if the
        movement succeeds. If the movement is not possible,
        the position will be unchanged and it will return False.
        """
        if not self._text_document.isValidPos(pos):
            return False

        self._pos = pos
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def isValid(self):
        return self._text_document.isValidPos(self._pos)

    def toLine(self, line_number):
        if not self.isValid():
            column = 1
        else:
            column = self.pos[1]

        if not self._text_document.isValidLine(line_number):
            return False

        self._pos = (line_number, min(self._text_document.lineLength(line_number), column))
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def toLineNext(self):
        next_line_number = self._pos[0]+1
        if not self._text_document.hasLine(next_line_number):
            return False

        self._pos = (next_line_number,
                     min(
                         self._text_document.lineLength(next_line_number),
                         max(self._pos[1],
                             self._optimistic_column
                        )
                      )
                    )
        self.positionChanged.emit(self._pos)
        return True

    def toLinePrev(self):
        prev_line_number = self._pos[0]-1
        if not self._text_document.hasLine(prev_line_number):
            return False

        self._pos = (prev_line_number,
                     min(
                         self._text_document.lineLength(prev_line_number),
                         max(self._pos[1],
                             self._optimistic_column
                         )
                     )
                   )
        self.positionChanged.emit(self._pos)
        return True

    def toCharNext(self):
        line_number = self._pos[0]
        next_char_number = self._pos[1]+1
        next_pos = (line_number, next_char_number)
        if not self._text_document.isValidPos(next_pos):
            return False

        self._pos = next_pos
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def toCharPrev(self):
        line_number = self._pos[0]
        prev_char_number = self._pos[1]-1
        prev_pos = (line_number, prev_char_number)

        if not self._text_document.isValidPos(prev_pos):
            return False

        self._pos = prev_pos
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def toLineBeginning(self):
        if not self.isValid():
            return False

        self._pos = (self._pos[0], 1)
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def toCharFirstNonBlank(self):
        """Moves the cursor along the line, to the first character that is not a blank (e.g. space)."""
        if not self.isValid():
            return False

        text = self._text_document.lineText(self._pos[0])
        lstrip_text = text.lstrip()
        if len(lstrip_text) == 0:
            self._pos = (self._pos[0], 1)
        else:
            self._pos = (self._pos[0], 1 + len(text) - len(lstrip_text))
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def toCharFirstNonBlankForLine(self, line_number):
        """Moves the cursor to the first non-blank character on the specified line"""
        if not self.isValid():
            return False

        text = self._text_document.lineText(line_number)
        lstrip_text = text.lstrip()
        if len(lstrip_text) == 0:
            self._pos = (line_number, 1)
        else:
            self._pos = (line_number, 1 + len(text) - len(lstrip_text))
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def toLineEnd(self):
        if not self.isValid():
            return False

        self._pos = (self._pos[0], self._text_document.lineLength(self._pos[0]))
        self._optimistic_column = self._pos[1]
        self.positionChanged.emit(self._pos)
        return True

    def toFirstLine(self):
        self._pos = (1,1)
        self._optimistic_column = 0
        self.positionChanged.emit(self._pos)
        return True

    def toLastLine(self):
        self._pos = (self._text_document.numLines(), 1)
        self._optimistic_column = 0
        self.positionChanged.emit(self._pos)
        return True




class BufferCommand(object):
    MEMENTO_INSERT, MEMENTO_REPLACE = list(range(2))
    def __init__(self, buffer):
        self._buffer = buffer
        self._document = buffer.document
        self._cursor = buffer.cursor

        self._line_memento_data = []
        self._sub_command = None
        self._saved_cursor_pos = None

    def saveCursorPos(self):
        self._saved_cursor_pos = self._cursor.pos

    def restoreCursorPos(self):
        self._cursor.toPos(self._saved_cursor_pos)

    def savedCursorPos(self):
        return self._saved_cursor_pos

    def saveLineMemento(self, line_number, restore_strategy):
        self._line_memento_data.append( (line_number, restore_strategy, self._document.lineMemento(line_number)) )

    def restoreLineMemento(self):
        line_number, restore_strategy, memento = self._line_memento_data.pop()

        if restore_strategy == BufferCommand.MEMENTO_INSERT:
            self._document.insertFromMemento(line_number, memento)
        elif restore_strategy == BufferCommand.MEMENTO_REPLACE:
            self._document.replaceFromMemento(line_number, memento)
        else:
            raise Exception("Unknown restore mode for memento")

    def lastSavedMemento(self):
        return self._line_memento_data[-1]

    def executeSubCommand(self, subcommand_type):
        sub_command = subcommand_type(self._buffer)
        result = sub_command.execute()
        if result.success:
            self._sub_command = sub_command
        
        return result
    
    def undo(self):
        if self._sub_command is not None:
            self._sub_command.undo()
            self._sub_command = None
            return

        for i in range(len(self._line_memento_data)):
            self.restoreLineMemento()

        if self._saved_cursor_pos is not None:
            self.restoreCursorPos()
